
typelist = ["int","long","double","float","double_complex","float_complex"]

complex_typelist = ["double_complex","float_complex"]

gemm_typelist = ["double","float","double_complex","float_complex"]

transform_typelist = [["double", "double"],
                      ["float", "float"],
                      ["double_complex", "double_complex"],
                      ["float_complex", "float_complex"],
                      ["double_complex", "double"],
                      ["double"        ,"double_complex"],
                      ["float_complex", "float"],
                      ["float"         ,"float_complex"]]

f = open("tensor_spec.h","w")
for t in typelist:
    f.write("\n// Instantiations for %s\n" % t)
    f.write("template class Tensor<%s>;\n" % t)
    f.write("template class SliceTensor<%s>;\n" % t)
    f.write("template std::ostream& operator << (std::ostream& s, const Tensor<%s>& t);\n" % t)
    f.write("template Tensor<%s> copy(const Tensor<%s>& t);\n" % (t,t))
    f.write("template Tensor<%s> outer(const Tensor<%s>& left, const Tensor<%s>& right);\n" % (t,t,t))
    f.write("template Tensor< Tensor<%s>::scalar_type > abs(const Tensor<%s>& t);\n" % (t,t))
    f.write("template Tensor<%s> transpose(const Tensor<%s>& t);\n" % (t,t))

f.write("\n// Instantiations for inner, inner_result and transform \n")
for t,q in transform_typelist:
    f.write("template void inner_result(const Tensor<%s>& left, const Tensor<%s>& right,\n" % (t,q))
    f.write("                           long k0, long k1, Tensor< TensorResultType<%s,%s>::type >& result);\n" % (t,q))
    f.write("template Tensor<TensorResultType<%s,%s>::type> inner(const Tensor<%s>& left, const Tensor<%s>& right,\n" % (t,q,t,q))
    f.write("                         long k0, long k1);\n")
    f.write("template Tensor<TensorResultType<%s,%s>::type> transform(const Tensor<%s>& t, const Tensor<%s>& c);\n" % (t,q,t,q))
    f.write("template Tensor<TensorResultType<%s,%s>::type> general_transform(const Tensor<%s>& t, const Tensor<%s> c[]);\n" % (t,q,t,q))
    f.write("template Tensor<TensorResultType<%s,%s>::type>& fast_transform(const Tensor<%s>& t, const Tensor<%s>& c, Tensor< TensorResultType<%s,%s>::type >& result, Tensor< TensorResultType<%s,%s>::type >& work);\n" % (t,q,t,q,t,q,t,q))


f.write("\n// Instantiations only for complex types\n")
for t in complex_typelist:
    f.write("\n// Instantiations for %s" % t)
    f.write("template Tensor< Tensor<%s>::scalar_type > arg(const Tensor<%s>& t);\n" % (t,t))
    f.write("template Tensor< Tensor<%s>::scalar_type > real(const Tensor<%s>& t);\n" % (t,t))
    f.write("template Tensor< Tensor<%s>::scalar_type > imag(const Tensor<%s>& t);\n" % (t,t))
    f.write("template Tensor<%s> conj(const Tensor<%s>& t);\n" % (t,t))
    f.write("template Tensor<%s> conj_transpose(const Tensor<%s>& t);\n" % (t,t))

#for t in typelist:
#    for q in typelist:
#        f.write("template TENSOR_RESULT_TYPE(%s,%s) Tensor<%s>::trace_conj(const Tensor<%s>& q);\n" % (t,q,t,q))

f.close()
    
## f.write("\n\nPUT THESE AT THE BOTTOM OF mxm.cc\n"
## for t in typelist:
##     f.write("template void mTxm<%s>(long dimi, long dimj, long dimk, %s * c, const %s *a, const %s *b);" % (t,t,t,t)
##     f.write("template void mxmT<%s>(long dimi, long dimj, long dimk, %s * c, const %s *a, const %s *b);" % (t,t,t,t)
##     f.write("template void mTxmT<%s>(long dimi, long dimj, long dimk, %s * c, const %s *a, const %s *b);" % (t,t,t,t)
##     f.write("template void mxm<%s>(long dimi, long dimj, long dimk, %s * c, const %s *a, const %s *b);" % (t,t,t,t)
    

##f.write("\n\nPUT THESE AT THE BOTTOM OF tensoriter.cc\n"
f = open("tensoriter_spec.h","w")

a = []
for t in typelist:
    for q in typelist:
        for r in typelist:
            a.append("template class TensorIterator<%s,%s,%s>;\n" % (t,q,r))
a.sort()

prev = a[0]
f.write(prev)
for cur in a[1:]:
    if cur != prev:
        f.write(cur)
        prev = cur
        
f.close()



import math
import mathutil
from array import *
from mathutil import printvector, printmatrix, dot
from longfloat import longfloat
from quadrature import Quadrature, QuadratureTest

# Compute the two-scale coefficients for Aplert's multi-wavelets
# This has been separated from the other code since we need extended
# precision to do this.
#
# Import this file using
#
# from twoscalecoeffs import twoscalecoeffs
#
# otherwise you'll end up with a bunch of multiple precision stuff that
# you really don't want

# Redefine a vector to be a list rather than an array of doubles
# so that we can use extended precision

def vector(n):
    return range(n)

def zerovector(n):
    v = vector(n)
    for i in range(n):
        v[i] = longfloat(0)
    return v

def zeromatrix(n,m): 
    a = range(n) 
    for i in range(n): 
        a[i] = zerovector(m) 
    return a 

def pn(x,order):
    '''
    Evaluate the Legendre polynomials up to the given order at x
    defined on [-1,1].
    '''
    p = range(order+1)
    p[0] = 1
    if order == 0:
        return p
    p[1] = x
    for n in range(1,order):
        p[n+1] = n*(x*p[n] - p[n-1])/(n+1) + x*p[n]
    return p

phi_norms = []                          # Cache sqrt(2*n+1)
def phi(x,k):
    '''
    Evaluate the shifted normalized Legendre polynomials up to the
    given order at x defined on [0,1].

    These are also our scaling functions, phi_i(x) , i=0..k-1

    In addition to forming an orthonormal basis on [0,1] we have
    phi_j(1/2-x) = (-1)^j phi_j(1/2+x)

    (the wavelets are similar with phase (-1)^(j+k)).
    '''
    global phi_norms

    if len(phi_norms) == 0:
        for n in range(max(k,100)):
            phi_norms.append(math.sqrt(2*n+1))
    order = k-1
    p = pn(2.*x-1,order)
    for n in range(k):
        p[n] = p[n]*phi_norms[n]  # sqrt(2*n+1)
    return p

def psi(x,k,g0,g1):
    '''
    For debug and demo purposes evaluate the multi-wavelets at x in [0,1]
    given the two-scale coefficients.
    '''
    if x < 0.5:
        p = phi(2.0*x,k)
        g = g0
    else:
        p = phi(2.0*x-1.0,k)
        g = g1

    root2 = math.sqrt(2.0)
    value = mathutil.zerovector(k)
    for i in range(k):
        for j in range(k):
            value[i] = value[i] + root2*g[i][j]*p[j]

    return value


def pnlong(x,order):
    '''
    USE LONGFLOATS
    
    Evaluate the Legendre polynomials up to the given order at x
    defined on [-1,1].
    '''
    p = range(order+1)
    p[0] = longfloat(1)
    x = longfloat(x)
    if order == 0:
        return p
    p[1] = x
    for n in range(1,order):
        p[n+1] = n*(x*p[n] - p[n-1])/(n+1) + x*p[n]
    return p

phi_norms_long = []                          # Cache sqrt(2*n+1)
def philong(x,k):
    '''
    USE LONGFLOATS
    
    Evaluate the shifted normalized Legendre polynomials up to the
    given order at x defined on [0,1].

    These are also our scaling functions, phi_i(x) , i=0..k-1

    In addition to forming an orthonormal basis on [0,1] we have
    phi_j(1/2-x) = (-1)^j phi_j(1/2+x)

    (the wavelets are similar with phase (-1)^(j+k)).
    '''
    global phi_norms_long

    if len(phi_norms_long) == 0:
        for n in range(max(k+1,61)):
            phi_norms_long.append(longfloat(2*n+1).sqrt())
    if k > 60:
        raise "phi_norms_long too small"
    
    order = k-1
    p = pnlong(2*x-1,order)
    for n in range(k):
        p[n] = p[n]*phi_norms_long[n]  # sqrt(2*n+1)
    return p

def h0h1(k):
    '''
    USE LONGFLOATS
    
    Compute the h0 & h1 two-scale coefficients

    h0_ij = sqrt(2)*int(phi_i(x)*phi_j(2x),x=0..1/2)
    h1_ij = sqrt(2)*int(phi_i(x)*phi_j(2x-1),x=1/2..1)

    but
    
    h1_ij = (-1)^(i+j) h0_ij from symmetry property of the phi

    The h0/h1 are useful because

    phi_i(x) = sqrt(2)*sum(j) h0_ij*phi_j(2x) + h1_ij*phi_j(2x-1)

    and
    
    phi_i(2x)   = (1/sqrt(2))*sum(j) h0_ji*phi_j(x) + g0_ji*psi(x)
    phi_i(2x-1) = (1/sqrt(2))*sum(j) h1_ji*phi_j(x) + g1_ji*psi(x)
    '''
    q = Quadrature(k,[longfloat(0),longfloat((-1,1))],uselongfloat=1)
    x = q.points()
    w = q.weights()
    npt = len(x)
    twox = range(npt)
    for i in range(npt):
        twox[i] = x[i] + x[i]

    h0 = zeromatrix(k,k)
    h1 = zeromatrix(k,k)
    
    root2 = longfloat(2).sqrt()
    for t in range(npt):
        p = philong(x[t],k)
        p2= philong(twox[t],k)
        for i in range(k):
            faci = p[i]*w[t]*root2
            for j in range(k):
                h0[i][j] = h0[i][j] + faci*p2[j]

    for i in range(k):
        for j in range(k):
            h1[i][j] = ((-1)**(i+j))*h0[i][j]

    return (h0, h1)

def orthonormalize_wavelets(g0,g1,k):
    '''
    USE LONGFLOATS
    
    Orthonormalize in reverse order
    '''
    for i in range(k-1,-1,-1):
        for attempt in range(1):   # was 2
            # Orthog to previous functions
            for m in range(k-1,i,-1):
                dotim = dot(g0[i]+g1[i],g0[m]+g1[m])
                for j in range(k):
                    g0[i][j] = g0[i][j] - dotim*g0[m][j]
                    g1[i][j] = g1[i][j] - dotim*g1[m][j]
            norm = dot(g0[i]+g1[i],g0[i]+g1[i])
            rnorm = norm.rsqrt()
            for j in range(k):
                g0[i][j] = g0[i][j]*rnorm
                g1[i][j] = g1[i][j]*rnorm

def half_moments(maxm,k):
    '''
    USE LONGFLOATS
    
    Compute the moments of the scaling functions computed
    over half over the interval.

    hmom1[p][j] = int(x^p * phi_j(2*x), x=0..1/2)
    hmom2[p][j] = int((x+0.5)^p * phi_j(2*x), x=0..1/2)

    hmom2 is made from hmom1 using the binomial expansion
    (x+a)^p = a^0*x^p + p*a^1*x^p-1 + p*(p-1)*a^2*x^p-2 / 2 + ... + a^p*x^0

    From Alpert, Beylkin, Gines and Vozovoi

    int(x^p phi_j(x),x=0..1) = sqrt(2*j+1)*(p!)^2 / ((p-j)!(p+j+1)!) p>=j

    Get an extra factor of (1/2)^(m+1) from the change of scale.

    For completeness also record here that
    
    int((x-1)^p phi_j(x),x=0..1) = (-1)^(j+p) *
    .                          sqrt(2*j+1)*(p!)^2 / ((p-j)!(p+j+1)!)

    Moments with p < j are zero due to orthogonality.
    '''

    hmom1 = zeromatrix(maxm+1,k)
    hmom2 = zeromatrix(maxm+1,k)

    factorial = range(maxm+1+k)
    pfac = longfloat(1)
    for p in range(maxm+1+k):
        factorial[p] = pfac
        pfac = pfac * (p+1)

    half = longfloat((-1,1))
    for p in range(maxm+1):
        pfac2 = factorial[p]*factorial[p]
        scale = longfloat(half)**(p+1)
        for j in range(k):
            if p < j:
                hmom1[p][j] = longfloat(0)
            else:
                hmom1[p][j] = scale*(longfloat(2*j+1).sqrt())*pfac2 / \
                             (factorial[p-j]*factorial[p+j+1])

    for p in range(maxm+1):
        for i in range(k):
            fac = longfloat(1)
            for q in range(p+1):
                hmom2[p][i] = hmom2[p][i] + fac*(half**(p-q))*hmom1[q][i]
                fac = fac * (p-q) / (q+1)
    return (hmom1, hmom2)
    

def g0g1(k):
    '''
    USE LONGFLOATS
    
    Compute the g0 & g1 two-scale coefficients which define the
    multi-wavelet in terms of the scaling functions at the finer
    level.

    psi_i(x) = sqrt(2)*sum(j) g0_ij*phi_j(2x) + g1_ij*phi_j(2x-1)

    Projection with sqrt(2)*phi_j(2x) and sqrt(2)*phi_j(2x-1), which
    are the normalized scaling functions on the finer level, yields 

    g0_ij = sqrt(2)*int(psi_i(x)*phi_j(2x),x=0..1/2)
    g1_ij = sqrt(2)*int(psi_i(x)*phi_j(2x-1),x=1/2..1)


    With the original Alpert conditions we have
    
    g1_ij = (-1)^(i+j+k) g0_ij from symmetry properties of the psi

    but you only get this if you impose the condition of additional
    vanishing moments.

    The projection of the additional moments is very numerically
    unstable and forces the use of higher precision for k >= 6.
    '''

    g0 = zeromatrix(k,k)
    g1 = zeromatrix(k,k)

    # Initialize the basis
    rroot2 = longfloat(2).rsqrt()
    for i in range(0,k):
        g0[i][i] =  rroot2
        g1[i][i] = -rroot2

    (h0, h1) = h0h1(k)

    # Project out the moments == orthog to the scaling functions.
    # Need to use the two-scale relation for the scaling functions.
    for i in range(k):
        for attempt in range(1): # was 2
            for m in range(k):
                dotim = dot(g0[i],h0[m])+dot(g1[i],h1[m])
                for j in range(k):
                    g0[i][j] = g0[i][j] - dotim*h0[m][j]
                    g1[i][j] = g1[i][j] - dotim*h1[m][j]

    # Orthog to x^k, x^k+1, ... , x^2k-2
    first = 0
    root2 = longfloat(2).sqrt()
    (hmom1,hmom2) = half_moments(2*k-2,k)

    test = longfloat(1).eps().sqrt()
    for power in range(k,2*k-1):
        # Transform to the wavelet basis
        overlap = zerovector(k)
        for i in range(k):
            for j in range(k):
                overlap[i] = overlap[i] + root2*\
                  (g0[i][j]*hmom1[power][j] + g1[i][j]*hmom2[power][j])

        # Put first function with non-zero overlap in front of the list
        for i in range(first,k):
            if abs(overlap[i]) > test:       # Is there a better test?
                (g0[first],g0[i]) = (g0[i],g0[first])
                (g1[first],g1[i]) = (g1[i],g1[first])
                (overlap[first],overlap[i]) = (overlap[i],overlap[first])
                break

        # Project out the first component as necessary.
        if abs(overlap[first]) > test:       # Is there a better test?
            for i in range(first+1,k):
                scale = overlap[i]/overlap[first]
                for j in range(k):
                    g0[i][j] = g0[i][j] - scale*g0[first][j]
                    g1[i][j] = g1[i][j] - scale*g1[first][j]

        first = first + 1
        
    orthonormalize_wavelets(g0,g1,k)
        
    return (g0, g1)

def fromfile(k):
    '''
    See if the two-scale coeffs have already been computed and stored on disk.
    Return the data are successfully read from disk, otherwise None.
    '''
    fname = '/tmp/two-scale-coeffs.' + str(k)
    try:
        f = open(fname,'rb')
        data = array('d',[])
        data.fromfile(f,4*k*k)
        h0 = mathutil.zeromatrix(k,k)
        h1 = mathutil.zeromatrix(k,k)
        g0 = mathutil.zeromatrix(k,k)
        g1 = mathutil.zeromatrix(k,k)
        pt = 0
        for i in range(k):
            for j in range(k):
                h0[i][j] = data[pt]
                h1[i][j] = data[pt+1]
                g0[i][j] = data[pt+2]
                g1[i][j] = data[pt+3]
                pt = pt + 4
        #print ' Read two-scale coefficients from file',fname
        return (h0,h1,g0,g1)
    except IOError, EOFError:
        return None

def tofile(h0,h1,g0,g1):
    '''
    Write the two-scale coeffs to disk.
    '''
    k = len(h0)
    fname = '/tmp/two-scale-coeffs.' + str(k)
    f = open(fname,'w+b')
    data = array('d',[0]*(4*k*k))
    pt = 0
    for i in range(k):
        for j in range(k):
            data[pt  ] = h0[i][j]
            data[pt+1] = h1[i][j]
            data[pt+2] = g0[i][j]
            data[pt+3] = g1[i][j]
            pt = pt + 4
    data.tofile(f)
    f.close()
    #print ' Wrote two-scale coefficients to file',fname


def twoscalecoeffs(k,usecached=1):
    '''
    Return the two-scale coefficients for Alperts basis of given order

    coarser <= finer
    phi_i(x) = sqrt(2)*sum(j) h0_ij*phi_j(2x) + h1_ij*phi_j(2x-1)
    ditto for psi_i(x) with g instead of h

    s_n,i,l = sum(j) h0_ij*s_n+1,j,2l + h1_ij*s_n+1,j,2l+1
    ditto for d with g instead of h

    finer <= coarser
    phi_i(2x)   = (1/sqrt(2))*sum(j) h0_ji*phi_j(x) + g0_ji*psi(x)
    phi_i(2x-1) = (1/sqrt(2))*sum(j) h1_ji*phi_j(x) + g1_ji*psi(x)

    s_n+1,i,2l   = sum(j) h0_ji*sn,j,l + g0_ji*d_n,j,l
    s_n+1,i,2l+1 = sum(j) h1_ji*sn,j,l + g1_ji*d_n,j,l

    '''
    global phi_norms_long

    if usecached:
        cached = fromfile(k)
        if cached:
            return cached

    nbitssave = longfloat.nbits
    if k < 11:
        longfloat.nbits = 156
    elif k < 19:
        longfloat.nbits = 208
    elif k < 24:
        longfloat.nbits = 260
    elif k < 29:
        longfloat.nbits = 320
    elif k < 33:
        longfloat.nbits = 360
    elif k < 37:
        longfloat.nbits = 400
    elif k < 42:
        longfloat.nbits = 440
    elif k < 46:
        longfloat.nbits = 480
    elif k < 49:
        longfloat.nbits = 520
    else:
        longfloat.nbits = 800

    longfloat.nbits = 800
    print ' Generating two-scale coeffcients for k=%d using %d-bit floating point numbers' \
          % (k, longfloat.nbits)
    
    # Force recomputation of the cached sqrt(2*j+1) values 
    phi_norms_long = []
    
    # Compute the coeffcicients to high precision using longfloats
    (h0l,h1l) = h0h1(k)
    (g0l,g1l) = g0g1(k)

    # Copy them into arrays of floats
    h0 = mathutil.zeromatrix(k,k)
    g0 = mathutil.zeromatrix(k,k)
    h1 = mathutil.zeromatrix(k,k)
    g1 = mathutil.zeromatrix(k,k)

    gerr = 0.0
    herr = 0.0
    for i in range(k):
        for j in range(k):
            hphase = (-1)**(i+j)
            gphase = (-1)**(i+j+k)
            h0[i][j] = float(h0l[i][j])
            h1[i][j] = h0[i][j]*hphase
            g0[i][j] = float(g0l[i][j])
            g1[i][j] = g0[i][j]*gphase
            if abs(h0[i][j]) < 4e-16: h0[i][j] = 0.0
            if abs(h1[i][j]) < 4e-16: h1[i][j] = 0.0
            if abs(g0[i][j]) < 4e-16: g0[i][j] = 0.0
            if abs(g1[i][j]) < 4e-16: g1[i][j] = 0.0
            testerr = float(abs(g0l[i][j]-gphase*g1l[i][j]))
            if g0[i][j] != 0.0: testerr = testerr/abs(g0[i][j])
            gerr = max(gerr,testerr)
            testerr = float(abs(h0l[i][j]-hphase*h1l[i][j]))
            if h0[i][j] != 0.0: testerr = testerr/abs(h0[i][j])
            herr = max(herr,testerr)
            
    print ' maximum symmetry error in h ', float(herr)
    print ' maximum symmetry error in g ', float(gerr)

    longfloat.nbits = nbitssave
    
    if max(float(herr),float(gerr)) > 4e-16:
        raise "twoscalecoeffs have inadequate precision"

    tofile(h0,h1,g0,g1)

    return (h0,h1,g0,g1)

if __name__ == "__main__":

    (h0,h1,g0,g1) = twoscalecoeffs(4)
    print h0
    print h1
    print g0
    print g1
    atop


    print '\n\n Beginning new run \n'

    #QuadratureTest(uselongfloat=1)

    for k in range(1,31):
        print '\n\n Testing multiwavelets k =', k, '\n'

        (h0,h1,g0,g1) = twoscalecoeffs(k)
        # Demonstrate that the scaling functions are ortho-normal
        i = 0
        j = 0
        def f(x):
            global k, i, j
            p = phi(x,k)
            return p[i]*p[j]
        q = Quadrature(k,[0,1.0])
        err = 0.0
        for i in range(k):
            for j in range(i+1):
                overlap = q.integrate(f)
                if i == j: overlap = overlap - 1
                err = max(err,abs(overlap))
        print ' Maximum errror in the overlap of scaling functions', err

        # Demonstrate that the wavelets are ortho-normal
        i = 0
        j = 0
        def f(x):
            global k, i, j, g0, g1
            p = psi(x,k,g0,g1)
            return p[i]*p[j]

        q1 = Quadrature(k,[0,0.5])
        q2 = Quadrature(k,[0.5,1.0])
        err = 0.0
        for i in range(k):
            for j in range(i+1):
                overlap = q1.integrate(f) + q2.integrate(f)
                if i == j: overlap = overlap - 1
                err = max(err,abs(overlap))
        print ' Maximum errror in the overlap of wavelets', err

        # Demonstrate that the wavelets are normal to scaling functions
        i = 0
        j = 0
        def f(x):
            global k, i, j, g0, g1
            p = psi(x,k,g0,g1)
            q = phi(x,k)
            return p[i]*q[j]

        q1 = Quadrature(k,[0,0.5])
        q2 = Quadrature(k,[0.5,1.0])
        err = 0.0
        for i in range(k):
            for j in range(k):
                overlap = q1.integrate(f) + q2.integrate(f)
                err = max(err,abs(overlap))
        print ' Maximum errror in the overlap of wavelets with scaling functions', err

        if k <= 4:
            print '     h0 '
            printmatrix(h0)
            print '     g0 '
            printmatrix(g0)
        



#!/usr/bin/python
import sys, math
#cart = False corresponds to spherical coordinates
cart = False
zaxis = False
eps = 0.001
if(cart):
  if(len(sys.argv) != 7):
    sys.exit("Requires arguments: xMIN xMAX dx zMIN zMAX dz");
if(not cart):
  if(len(sys.argv) != 5):
    sys.exit("Requires arguments: kMIN dk kMAX   nTH=PI/nTH");
def floatRange(a, b, inc):
  """
  Returns a list containing an arithmetic progression of floats.
  This is simply a float version of the built-in range(a, b, step)
  function.  The result is [ , ) as always in Python.
  """
  try: x = [float(a)]
  except: return False
  for i in range(1, int(math.ceil((b - a ) / inc))):
    x. append(a + i * inc)
  return x

if(cart):
  xMIN = float(sys.argv[1])
  xMAX = float(sys.argv[2])
  dx = float(sys.argv[3])
  zMIN = float(sys.argv[4])
  zMAX = float(sys.argv[5])
  dz = float(sys.argv[6])
  for x in floatRange(xMIN, xMAX+eps, dx):
    for z in floatRange(zMIN, zMAX+eps, dz):
      #if(math.sqrt(x*x + z*z)<min(xMAX,zMAX)):
      print "0 %3.1f %3.1f" % (x, z)
else:
  kMIN = float(sys.argv[1])
  dk = float(sys.argv[2])
  kMAX = float(sys.argv[3])
  PI = 3.14159265358979
  dTH = PI/float(sys.argv[4])
  for k in floatRange(kMIN, kMAX+eps, dk):
    if( math.fabs(k)<eps ):
      break
    for th in floatRange(0, PI+eps, dTH):
      print "0 %13.12f %13.12f" % (k*math.sin(th), k*math.cos(th))

# z-axis resolution (this messes up numerical integration)
# Add resolution to the z-oaxis
if(zaxis):
#  print "k%dk  "
  for k in floatRange(kMIN, kMAX+eps, dk/5):
#    print math.fabs((k + eps/100)%dk), "   "
    if( math.fabs((k + eps/100)%dk) > eps ):
      print "0 0 %5.4f" % k





# The stuff in this file is now obsolete and is
# here to keep the old twoscalecoeff code working.

# Don't write new code using this stuff.


from array import * 
from math import sqrt 

def vector(n):
    return array('d',[0.0]*n)

def zerovector(n): 
    return vector(n)

def zeromatrix(n,m): 
    a = range(n) 
    for i in range(n): 
        a[i] = zerovector(m) 
    return a 

def copyvector(x):
    n = len(x)
    a = vector(n)
    for i in range(n): 
        a[i] = x[i] 
    return a 

def copymatrix(x): 
    n = len(x) 
    m = len(x[0]) 
    a = zeromatrix(n,m) 
    for i in range(n): 
        for j in range(m): 
            a[i][j] = x[i][j] 
    return a 

def transpose(x): 
    n = len(x) 
    m = len(x[0]) 
    a = zeromatrix(m,n) 
    for i in range(n): 
        for j in range(m): 
            a[j][i] = x[i][j] 
    return a 

def dot(a,b): 
    sum = 0
    for i in range(len(a)): 
        sum = sum + a[i]*b[i] 
    return sum 

def vector_norm(x):
    '''
    A translation of the BLAS routine dnrm2
    '''
    n = len(x)
    if n < 1:
        return 0.0
    elif n == 1:
        return abs(x[0])
    else:
        scale = 0.0
        ssq = 1.0
        for i in range(n):
            if x[i] != 0.0:
                absxi = abs(x[i])
                if scale < absxi:
                    ssq = 1.0 + ssq*(scale/absxi)**2
                    scale = absxi
                else:
                    ssq = ssq + (absxi/scale)**2
        return scale*sqrt(ssq)

def mxm(a,b): 
    n = len(a) 
    k = len(a[0]) 
    kk= len(b) 
    m = len(b[0]) 
    if kk != k: 
        raise "matrices do not conform for multiplication" 
    c = zeromatrix(n,m) 
    for i in range(n): 
        for j in range(m): 
            sum = 0.0 
            for l in range(k): 
                sum = sum + a[i][l]*b[l][j] 
            c[i][j] = sum 
    return c

sparse_mtxm_nflop = 0.0
def sparse_mtxm(a,b):
    global sparse_mtxm_nflop
    '''
    C = AT * B using sparsity in both.  Currently assumes full storage.
    '''
    k = len(a) 
    kk= len(b) 
    n = len(a[0]) 
    m = len(b[0]) 
    if kk != k: 
        raise "matrices do not conform for multiplication" 
    ct = zeromatrix(m,n)  # This is the transpose of the result
    nflop = 0.0
    for l in range(k):
        al = a[l]
        bl = b[l]
        list = [] # List of non-zero a[l][i] for given l
        for i in range(n):
            if al[i]: list.append(i)
        numl = float(len(list))
        for j in range(m):
            blj = bl[j]
            if blj:
                ctj = ct[j]
                nflop = nflop + numl
                for i in list:
                    ctj[i] = ctj[i] + al[i]*blj
    sparse_mtxm_nflop = sparse_mtxm_nflop + nflop
    return transpose(ct)

def mxv(a,b): 
    n = len(a) 
    k = len(a[0]) 
    kk = len(b) 
    if k != kk: 
        raise "matrix and vector do not conform for multiplication" 
    c = zerovector(n) 
    for i in range(n): 
        c[i] = dot(a[i],b) 
    return c 

def printvector(a): 
    n = len(a)
    for i in range(n): 
        print ("%+15.8e "%float(a[i])),
        if ((i+1)%5) == 0 and i != (n-1):
            print " "
    print " " 

def printmatrix(a): 
    n = len(a)
    m = len(a[0])
    for jlo in range(0,m,5):
        print "    ",
        for j in range(jlo,min(jlo+5,m)):
            print ("     %5i       " % j),
        print " "
        for i in range(n):
            print ("%4i  " % i),
            for j in range(jlo,min(jlo+5,m)):
                print ("%+15.8e " % float(a[i][j])),
            print " "

def numderiv(func,x,step,eps): 
    ''' 
    Use central differences to compute the gradient and diagonal 
    elements of the Hessian. 
    func(x) = function to be differentiated 
    x[] = point at which to differentiate 
    step[] = remembers finite difference step between 
    .        successive calls.  Set to zero on first call 
    .        or set close to appropriate value 
    eps = expected precision in func 
    
    Some care is taken to adjust the step so that the gradient and 
    Hessian diagonal are estimated with about 4 digits of precision 
    but some noise is unavaoidable due either to the noise in the 
    function or cubic/higher terms in the Taylor expansion. 
    ''' 
    
    n = len(x) 
    g = zerovector(n) 
    h = zerovector(n) 
    f0 = func(x) 
    for i in range(n): 
        if step[i] == 0.0: 
            step[i] = max(abs(x[i])*0.01,0.0001) 
        xi = x[i] 
        while 1: 
            x[i] = xi + step[i] 
            f1 = func(x)
            measure = 1e4*eps*max(abs(f0),eps)
            if abs(f1-f0) < measure:
                #print ' Increasing step ',i,step[i],abs(f1-f0) 
                step[i] = step[i]*2.0 
            elif abs(f1-f0) > 10.0*measure:
                #print ' Decreasing step ',i,step[i],abs(f1-f0) 
                step[i] = step[i]/3.0 
            else: 
                break 
        x[i] = xi - step[i] 
        fm1 = func(x) 
        x[i] = xi 
        g[i] = (f1 - fm1)/(2*step[i]) 
        h[i] = (f1 + fm1 - 2.0*f0)/(step[i]*step[i]) 

    return (f0,g,h) 



def quadfit(alpha0, f0, alpha1, f1, alpha2, f2): 
    ''' 
    Given 3 points compute the gradient and hessian at point 0 
    using a quadratic fit. 
    ''' 
    delta1 = alpha1 - alpha0 
    delta2 = alpha2 - alpha0 
    d1 = (f1 - f0)/delta1 
    d2 = (f2 - f0)/delta2 
    h0 = 2.0*(d1 - d2)/(delta1-delta2) 
    g0 = d1 - 0.5*h0*delta1 

    test1 = f0 + g0*delta1 + 0.5*h0*delta1*delta1 
    test2 = f0 + g0*delta2 + 0.5*h0*delta2*delta2 
    return (f0, g0, h0) 

def takestep(x0, s, alpha): 
    x = zerovector(len(x0)) 
    for j in range(len(x)): 
        x[j] = x0[j] + s[j]*alpha 
    return x 

def quadratic_step(trust, g0, h0, prnt=1): 
    if h0 > 0: 
        delta2 = -g0/h0 
        if abs(delta2) > trust:
            if prnt:
                print "                    Step restriction: %f %f " % (delta2, trust) 
            delta2 = abs(trust*delta2)/delta2 
    else:
        if prnt:
            print "                    Negative curvature " 
        delta2 = -abs(trust*g0)/g0 
    return delta2 



def linesearch(func, x0, s, lsgrad, eps, prnt=1): 
    # Assume here that some halfway OK preconditioning 
    # is being used so we expect a step around unity. 
    # Nevertheless, must exercise some caution. 

    # First step in small increments until we've either 
    # bracketed the minimum or gone downhil with enough 
    # energy difference to start fitting

    if prnt:
        print " Line search: step   alpha     grad     hess        value" 
        print "              ---- --------- -------- --------  ----------------" 

    trust = 0.2 
    alpha0 = 0.0 
    f0 = func(x0)
    if prnt:
        print "                   %9.2e %8.1e          %16.8e" % \
              (alpha0, lsgrad, f0) 

    if lsgrad < 0: 
        alpha1 = alpha0 + trust 
    else: 
        alpha1 = alpha0 - trust 
    f1 = func(takestep(x0,s,alpha1))
    if prnt:
        print "                   %9.2e                   %16.8e" % \
              (alpha1, f1) 

    snorm = vector_norm(s)
    
    while f1 > f0: 
        if trust*snorm < sqrt(eps):
            if prnt:
                print " system is too badly conditioned for initial step" 
            return (alpha0,f0)          # Cannot seem to find my way 
        trust = trust * 0.5 
        if lsgrad < 0: 
            alpha1 = alpha0 + trust 
        else: 
            alpha1 = alpha0 - trust 
        f1 = func(takestep(x0,s,alpha1))
        if prnt:
            print "                   %9.2e                   %16.8e" % \
                  (alpha1, f1) 
        
    g0 = lsgrad 
    h0 = (f1-f0-alpha1*g0)/alpha1**2 
    if f1 < f0: 
        g0 = g0 + h0*(alpha1 - alpha0) 
        alpha0, alpha1, f0, f1 = alpha1, alpha0, f1, f0 
    
    alpha2 = alpha0 + quadratic_step(trust,g0,h0,prnt) 

    nbackstep =0 

    for iter in range(1,10): 
        f2 = func(takestep(x0,s,alpha2)) 
        #print ' alphas ', alpha0, alpha1, alpha2 
        #print ' fs     ', f0, f1, f2 
        
        if iter == 1: 
            f2prev = f2 
            
        # Check for convergence or insufficient precision to proceed further 
        if (abs(f0-f1)<(10*eps)) and (abs(f1-f2)<(10*eps)):
            if prnt:
                print "                  ", 
                print " Insufficient precision ... terminating LS" 
            break 
        
        if (f2-f2prev) > 0: 
            # New point is higher than previous worst 
            if  nbackstep < 3: 
                nbackstep = nbackstep + 1
                if prnt:
                    print "                  ", 
                    print " Back stepping due to uphill step" 
                trust = max(0.01,0.2*abs(alpha2 - alpha0))  # Reduce trust radius 
                alpha2 = alpha0 + 0.2*(alpha2 - alpha0) 
                continue 
        elif (f2-f0) < 0: 
            trust = min(4.0,trust*2.0)  # Seem to have narrowed the search 

        nbackstep = 0 

        f2prev = f2 

        # Order in increasing energy 
        if f1 < f0: 
            alpha0, alpha1, f0, f1 = alpha1, alpha0, f1, f0 
        if f2 < f0: 
            alpha0, alpha2, f0, f2 = alpha2, alpha0, f2, f0 
        if f2 < f1: 
            alpha1, alpha2, f1, f2 = alpha2, alpha1, f2, f1 

        (f0, g0, h0) = quadfit(alpha0, f0, alpha1, f1, alpha2, f2)
        if prnt:
            print "              %4i %9.2e %8.1e %8.1e %16.8e" % \
                  (iter, alpha0, g0, h0, f0) 

        if (h0>0.0) and (abs(g0) < 0.03*abs(lsgrad)):
            if prnt:
                print "                  ", 
                print " gradient reduced 30-fold ... terminating LS" 
            break 
        
        # Determine the next step 
        delta = quadratic_step(trust,g0,h0,prnt) 
        alpha2 = alpha0 + delta 
        df = g0*delta + 0.5*h0*delta*delta 
        if abs(df) < 10.0*eps:
            if prnt:
                print "                  ", 
                print " projected energy reduction < 10*eps ... terminating LS" 
            break 

    return (alpha0, f0) 

def jacobi(a): 
    ''' 
    Diagonalize a real symmetric matrix using the variable threshold 
    cyclic Jacobi method.   The input matrix is unmodified.
    
    (v,e) = jacobi(a) 
    
    Input: a[n][m] is a real symmetric matrix 

    Returns: (v,e) where v is the list of eigenvectors and e is an 
    vector of the corresponding eigenvalues in ascending order. 
    v[k] is a vector containing the kth eigenvector.  These satisfy 

    A*Vt = Vt*e 

    or 

    V*A = e*V 

    or 
    
    sum(j)(a[i][j]v[k][j]) = e[k]*v[k][i] 
    ''' 
    a = copymatrix(a) 
    n = len(a) 
    m = len(a[0]) 
    if n != m: 
        raise 'Matrix must be square' 
    for i in range(n): 
        for j in range(m): 
            if a[i][j] != a[j][i]: 
                raise ' Matrix must be symmetric' 

    tolmin = 5.0e-16 
    tol = 1e-2 

    v = zeromatrix(n,n) 
    for i in range(n): 
        v[i][i] = 1.0 

    maxd = 0.0 
    for i in range(n): 
        maxd = max(abs(a[i][i]),maxd) 

    nrotsum = 0
    for iter in range(50):
        maxdaij = 0.0
        nrot = 0 
        for i in range(n): 
            for j in range(i+1,n):  # j>i
                aii = a[i][i] 
                ajj = a[j][j]
                aij = a[i][j]
                daij = abs(aij)
                maxdaij = max(maxdaij,daij/maxd)
                if daij > tol*maxd: # Screen small elements 
                    s = ajj - aii 
                    ds = abs(s) 
                    if daij > (tolmin*ds): # Check for sufficient precision
                        nrot = nrot + 1 
                        if (tolmin*daij) > ds: 
                            c = s = 1/sqrt(2.) 
                        else: 
                            t = aij/s 
                            u = 0.25/sqrt(0.25+t*t) 
                            c = sqrt(0.5+u) 
                            s = 2.*t*u/c 

                        for k in range(i+1): 
                            t = a[k][i] 
                            u = a[k][j] 
                            a[k][i] = c*t - s*u 
                            a[k][j] = s*t + c*u
                            
                        ai = a[i]
                        aj = a[j]
                        for k in range(i+1,j): 
                            t = ai[k] 
                            u = a[k][j] 
                            ai[k] = c*t - s*u 
                            a[k][j] = s*t + c*u

                        a[j][j] = s*aij + c*ajj
                        a[i][i] = c*a[i][i] - s*(c*aij - s*ajj)

                        for k in range(j,n): 
                            t = ai[k] 
                            u = aj[k] 
                            ai[k] = c*t - s*u 
                            aj[k] = s*t + c*u
 
                        vi = v[i]
                        vj = v[j]
                        for k in range(n): 
                            t = vi[k] 
                            u = vj[k] 
                            vi[k] = c*t - s*u 
                            vj[k] = s*t + c*u 

                        a[j][i] = a[i][j] = 0.0 
                        maxd = max(maxd,abs(a[i][i]),abs(a[j][j]))
        #print iter, tol, maxdaij, nrot, '!'
        nrotsum = nrotsum + nrot
        if nrot == 0 and tol <= tolmin: 
            break
        tol = min(tol,maxdaij*1e-1,maxdaij**2)
        tol = max(tol,tolmin)

    #print "nrotsum", nrotsum
    if nrot != 0: 
        raise "Jacobi iteration did not converge in 50 passes"

    # Sort eigenvectors and values into increasing order 
    e = zerovector(n) 
    for i in range(n): 
        e[i] = a[i][i] 
        for j in range(i): 
            if e[j] > e[i]: 
                (e[i],e[j]) = (e[j],e[i]) 
                (v[i],v[j]) = (v[j],v[i]) 

    return (v,e) 

def hessian_update_bfgs(hp, dx, g, gp): 
    ''' 
    Apply the BFGS update to the approximate Hessian h[][]. 

    hp[][] = Hessian matrix from previous iteration 
    dx[]  = Step from previous iteration 
    .       (dx[] = x[] - xp[] where xp[] is the previous point) 
    g[]   = gradient at current point 
    gp[]  = gradient at previous point 

    Returns the updated hessian 
    ''' 

    n = len(hp) 
    hdx  = mxv(hp,dx) 
    dg = zerovector(n) 
    for i in range(n): 
        dg[i] = g[i] - gp[i] 
    
    dxhdx = dot(dx,hdx) 
    dxdx  = dot(dx,dx) 
    dxdg  = dot(dx,dg) 
    dgdg  = dot(dg,dg) 
    h = copymatrix(hp) 

    if (dxdx > 0.0) and (dgdg > 0.0) and (abs(dxdg/sqrt(dxdx*dgdg)) > 1.e-8): 
        for i in range(n): 
            for j in range(n): 
                h[i][j] = h[i][j] + dg[i]*dg[j]/dxdg - hdx[i]*hdx[j]/dxhdx 
    else: 
        print ' BFGS not updating dxdg (%e), dgdg (%e), dxhdx (%f), dxdx(%e)' % (dxdg, dgdg, dxhdx, dxdx) 

    return h 

def quasinr(func, guess, tol, eps, printvar=None, prnt=1): 
    ''' 
    Unconstrained minimization of a function of n variables 
    without analytic derivatives using quasi-Newtwon with BFGS update 
    and numerical gradients. 
    
    func(x) is a function that takes an array of n values and 
    returns the function value 
    
    guess[] is a vector of n values for the initial guess 
    
    tol is the convergence criterion for the maximum value 
    of the gradient 
    
    eps is the expected precision in the function value 
    
    printvar(x) is an optional user function to print the values of 
    parameters each macro iteration 
    ''' 
    
    n = len(guess) 
    x = copyvector(guess) 
    s = zerovector(n) 
    g = zerovector(n) 
    gp = zerovector(n) 
    step = zerovector(n) 
    hessian = zeromatrix(n,n) 

    alpha = 0.0 

    for iter in range(50*n): 
        (value,g,h) = numderiv(func, x, step, eps) 
        gnrm = vector_norm(g)
        
        if prnt:
            print ' ' 
            print ' iter    gnrm         value ' 
            print ' ---- --------- ----------------' 
            print "%4i %9.2e %16.8e" % (iter,gnrm,value) 

        if (prnt and printvar): 
            printvar(x) 

        if gnrm < tol: 
            if prnt: print ' Converged!' 
            break 

        if iter == 0: 
            for i in range(n): 
                hessian[i][i] = max(abs(h[i]),1e-4) 
        else: 
            hessian = hessian_update_bfgs(hessian, s, g, gp) 

        (v,e) = jacobi(hessian) 
        emax = max(map(abs,e)) 
        emin = emax*eps*100.0   # Control noise in small eigenvalues
        if prnt:
            print '\n Eigenvalues of the Hessian:' 
            printvector(e) 

        # Transform to spectral form, take step, transform back 
        gs = mxv(v,g) 
        for i in range(n): 
            if e[i] < emin:
                if prnt:
                    print ' Mode %d: small/neg eigenvalue (%f).' % (i, e[i]) 
                s[i] = -gs[i]/emin 
            else: 
                s[i] = -gs[i]/e[i] 
        s = mxv(transpose(v),s) 

        # Apply overall step restriction ... better LS obviates this 
        scale = 1.0 
        for i in range(n): 
            trust = abs(x[i])
            if trust == 0.0: trust = 0.1 # ????
            if abs(s[i]) > trust:
                if prnt:
                    print ' restricting ', i, trust, abs(x[i]), \
                          abs(x[i]/sqrt(abs(hessian[i][i]))), s[i] 
                scale = min(scale,trust/abs(s[i])) 

        if scale != 1.0: 
            for i in range(n): 
                s[i] = s[i]*scale 
        
        (alpha,value) = linesearch(func, x, s, dot(s,g), eps, prnt) 

        if alpha == 0.0:
            if prnt:
                print ' Insufficient precision to proceed further' 
            break 

        for i in range(n): 
            s[i] = s[i]*alpha 
            x[i] = x[i] + s[i] 
            gp[i]= g[i] 
    return (gnrm,x) 


def cgminold(func, dfunc, guess, tol): 
    ''' 
    Simple conjugate gradient assuming analtyic derivatives. 
    ''' 
    n = len(guess) 
    x = copyvector(guess) 
    s = zerovector(n) 
    g = zerovector(n) 
    gp= zerovector(n) 
    value = func(x) 

    for iter in range(10*n): 
        g = dfunc(x) 
        gmax = max(map(abs,g)) 
        
        print ' ' 
        print ' iter    gmax         value ' 
        print ' ---- --------- ----------------' 
        print "%4i %9.2e %16.8f" % (iter,gmax,value) 
        
        if gmax < tol: 
            print ' Converged!' 
            break 
        
        if (iter == 0) or ((iter%20) == 0): 
            beta = 0.0 
        else: 
            beta = (dot(g,g) - dot(g,gp))/(dot(s,g)-dot(s,gp)) 

        for i in range(n): 
            s[i] = -g[i] + beta*s[i] 

        (alpha,value) = linesearch(func, x, s, dot(s,g), 1e-12) 

        for i in range(n): 
            s[i] = s[i]*alpha 
            x[i] = x[i] + s[i] 
            gp[i]= g[i] 

    return (value,x) 



def cgmin(func, dfunc, guess, tol, precond=None, reset=None): 
    ''' 
    Conjugate gradient with optional preconditioning and 
    use of analytic gradients. 
    ''' 
    n = len(guess) 
    x = copyvector(guess) 
    s = zerovector(n) 
    g = zerovector(n) 
    gp= zerovector(n) 
    value = func(x) 

    if not reset: 
        reset = n 
    reset = min(reset,n) 

    for iter in range(10*n): 
        g = dfunc(x) 
        gmax = max(map(abs,g)) 
        
        print ' ' 
        print ' iter    gmax         value ' 
        print ' ---- --------- ----------------' 
        print "%4i %9.2e %16.8f" % (iter,gmax,value) 

        if gmax < tol: 
            print ' Converged!' 
            break 
        
        if precond: 
            precondg = precond(x,g) 
        else: 
            precondg = g 
        
        if (iter % reset) == 0: 
            beta = 0.0 
        else: 
            beta = (dot(precondg,g) - dot(precondg,gp))/(dot(s,g)-dot(s,gp)) 

        for i in range(n): 
            s[i] = -precondg[i] + beta*s[i] 

        (alpha,value) = linesearch(func, x, s, dot(s,g), 
                                   max(1e-16,abs(value)*1e-12)) 

        for i in range(n): 
            s[i] = s[i]*alpha 
            x[i] = x[i] + s[i] 
            gp[i]= g[i] 
        
    return (value,x) 

def cgmin2(func, guess, tol, eps, printvar=None,reset=None): 
    ''' 
    Unconstrained minimization of a function of n variables 
    without analytic derivatives using conjugate gradient with 
    diagonal preconditioning. 
    
    func(x) is a function that takes a vector of n values and 
    returns the function value 
    
    guess[] is a vector of n values for the initial guess 
    
    tol is the convergence criterion for the maximum value 
    of the gradient 
    
    eps is the expected precision in the function value 
    
    printvar(x) is an optional user function to print the values of 
    parameters each iteration 
    
    reset is the number of iterations between forced resets of the 
    conjugacy.  In principle this could be n but noise in the 
    numerical gradients makes a smaller number a better choice. 
    ''' 
    
    n = len(guess) 
    x = copyvector(guess) 
    s = zerovector(n) 
    g = zerovector(n) 
    gp = zerovector(n) 
    step = zerovector(n) 
    precondg = zerovector(n) 

    alpha = 0.0 

    if not reset: 
        reset = n 
    reset = min(reset,n) 

    for iter in range(50*n): 
        (value,g,hh) = numderiv(func, x, step, eps) 
        gmax = max(map(abs,g)) 
        
        print ' ' 
        print ' iter    gmax         value ' 
        print ' ---- --------- ----------------' 
        print "%4i %9.2e %16.8f" % (iter,gmax,value) 

        if (printvar): 
            printvar(x) 

        if gmax < tol: 
            print ' Converged!' 
            break 

        if (iter % reset) == 0: 
            # On the first iteration or if not applying conjugacy 
            # we can recompute the diagonal preconditioner 
            h = copyvector(hh) 
            for i in range(n): 
                h[i] = max(abs(h[i]),1e-6) 
        
        # Preconditioning with the diagonal of the Hessian 
        for i in range(n): 
            precondg[i] = g[i] / h[i] 
            
        # Should be able to reset every n steps but noisy gradients 
        # means that we don't have enough info. 
        if (iter % reset) == 0: 
            if iter != 0: 
                print" Resetting conjugacy" 
            beta = 0.0 
        else: 
            beta = (dot(precondg,g) - dot(precondg,gp))/(dot(s,g)-dot(s,gp)) 

        for i in range(n): 
            s[i] = -precondg[i] + beta*s[i] 

        (alpha,value) = linesearch(func, x, s, dot(s,g), eps) 

        if alpha == 0.0: 
            # LS failed, probably due to lack of precision. 
            if beta != 0.0: 
                print "LS failed - trying preconditioned steepest descent direction" 
                for i in range(n): 
                    s[i] = -g[i] 
                (alpha,value) = linesearch(func, x, s, dot(s,g), eps) 
            if alpha == 0.0: 
                print " Insufficient precision to proceed further" 
                break 

        for i in range(n): 
            s[i] = s[i]*alpha 
            x[i] = x[i] + s[i] 
            gp[i]= g[i] 
    return (value,x) 
    
def choleski(A):
    '''
    Returns the Choleski factorization of a square positive definite
    symmetric matrix.  Lij i<=j and A=L.Lt
    Raises an exception if A is not symmetric or is singular.
    '''
    n = len(A)
    m = len(A[0])
    if n != m:
        raise ValueError,"choleski factorization requires square matrix"
    for i in range(n):
        for j in range(n):
            aij = A[i][j]
            aji = A[j][i]
            if abs(aij - aji) > (1e-15*max(abs(aij),abs(aji))):
                raise ValueError,"choleski factorization requires symmetric matrix"
    L = zeromatrix(n,n)
    for k in range(n):
        sum = 0.0
        for m in range(k):
            sum = sum + L[k][m]*L[k][m]
        if A[k][k] <= sum:
            raise ValueError, "choleski factorization requires positive definite matrix"
        L[k][k] = sqrt(A[k][k]-sum)
        for i in range(k+1,n):
            sum = 0.0
            for m in range(k):
                sum = sum + L[i][m]*L[k][m]
            L[k][i] = L[i][k] = (A[i][k] - sum)/L[k][k]
    return L

def forward_elimination(L,b):
    '''
    In solving LUx=b, the first step is the forward elimination, y=L^-1.b
    '''
    n = len(L)
    y = zerovector(n)
    y[n-1] = b[n-1]/L[n-1][n-1]
    for i in range(n):
        sum = 0.0
        for j in range(i):
            sum = sum + L[i][j]*y[j]
        y[i] = (b[i] - sum)/L[i][i]
    return y

def backward_substitution(U,y):
    '''
    In solving LUx=b, the second step is the backward substitution, x=U^-1y
    '''
    n = len(U)
    x = zerovector(n)
    x[n-1] = y[n-1]/U[n-1][n-1]
    for i in range(n-2,-1,-1):
        sum = 0.0
        for j in range(i+1,n):
            sum = sum + U[i][j]*x[j]
        x[i] = (y[i] - sum)/U[i][i]
    return x

def choleski_solve(A,b):
    '''
    Solve Ax=b using Choleski factorization of A.
    '''
    L = choleski(A)
    y = forward_elimination(L,b)
    return backward_substitution(L,y)

def davidson(A,thresh=1e-6,maxsub=10,guess=None):
    '''
    Return the lowest eigenvalue and corresponding vector of A
    '''
    n = len(A)
    
    if guess:
        x = [copyvector(guess)]
        scale = 1.0/sqrt(vector_norm(x[0]))
        for i in range(n):
            x[0][i] = x[0][i]*scale
    else:
        x = [zerovector(n)]
        imin = 0
        for i in range(1,n):
            if A[i] < A[imin]:
                imin = i
        x[0][imin] = 1.0

    maxsub = min(maxsub,n)
    Ax = []
    nsub = 1
    for iter in range(10*n):
        Ax.append(mxv(A, x[nsub-1]))
        xAx = mxm(x,transpose(Ax))
        for j in range(nsub-1):
            for k in range(j+1,nsub):
                xAx[k][j] = xAx[j][k]
        #print 'Reduced matrix'
        #printmatrix(xAx)
        (v,e) = jacobi(xAx)
        #print 'Reduced evecs'
        #printmatrix(v)
        #print 'Reduced evals'
        #printvector(e)

        z = zerovector(n)
        bestx = zerovector(n)
        err = 0.0
        for i in range(n):
            Axi = 0.0
            xi  = 0.0
            for j in range(nsub):
                xi = xi + v[0][j]* x[j][i]
                Axi=Axi + v[0][j]*Ax[j][i]
            denom = (A[i][i] - e[0])
            if (denom < 0.01):
                denom = 0.01
            step = -(Axi - e[0]*xi)/denom
            #print i, step, Axi, e[0], xi, denom
            err = err + step*step
            z[i] = xi + step
            bestx[i] = xi
        scale = 1.0/sqrt(vector_norm(bestx))
        for i in range(n):
            bestx[i] = bestx[i]*scale
        print "%5d %5d %20.10f %9.1e" % (iter, nsub, e[0], sqrt(err))
        if sqrt(err) < thresh:
            return (bestx,e[0])
        
        for loop in range(2):        
            for j in range(nsub):
                zj = dot(z,x[j])
                for i in range(n):
                    z[i] = z[i] - zj*x[j][i]
        scale = 1.0/vector_norm(z)
        for i in range(n):
            z[i] = z[i]*scale
        if nsub < maxsub:
            x.append(z)
            nsub = nsub + 1
        else:
            nsub = 1
            x = [bestx]
            Ax = []
    raise "davidson did not converge"

if __name__ == '__main__':

    def precond(g): 
        # Used to test optional preconditioner for cgmin(). 
        precondg = copyvector(g) 
        for i in range(len(g)): 
            precondg[i] = precondg[i]/(i+2.0) 
        return precondg 

    def df(x): 
        d = zerovector(len(x)) 
        for i in range(len(x)): 
            d[i] = x[i]*(i+1) 
            for j in range(len(x)): 
                d[i] = d[i] + x[j] 
        return d 

    def f(x): 
        sum = 0.0 
        for i in range(len(x)): 
            for j in range(len(x)): 
                sum = sum + 0.5*x[i]*x[j] 
        for i in range(len(x)): 
            sum = sum + 0.5*x[i]*x[i]*(i+1) 
        return sum

    import random

    print '\n\n   TESTING DAVIDSON DIAGONALIZATION'
    n = 8
    a = zeromatrix(n,n)
    for i in range(n):
        for j in range(i,n):
            a[j][i] = a[i][j] = random.random()-0.5
        a[i][i] = a[i][i]*n

    (v,e) = jacobi(a)
    print ' Eval from jacobi', e[0]
    davidson(a)
    sys.exit()
    
    print '\n\n   TESTING SPARSE MATRIX PRODUCT'
    n = 13
    m = 17
    k = 11
    a = zeromatrix(k,n)
    b = zeromatrix(k,m)
    for i in range(k):
        for j in range(n):
            value = random.random()
            if (value < 0.25): a[i][j] = value
        for j in range(m):
            value = random.random()
            if (value < 0.25): b[i][j] = value
    cc= mxm(transpose(a),b)
    c = sparse_mtxm(a,b)
    err = 0.0
    for i in range(n):
        for j in range(m):
            err = err + abs(c[i][j]-cc[i][j])
    print "sparse flops:", sparse_mtxm_nflop, "dense flops:", n*m*k
    if (err < 1e-14):
        print "  OK"
    else:
        print " FAILED"
    import sys
    sys.exit()

    print '\n\n   TESTING JACOBI EIGENSOLVER\n\n' 
    n = 5
    a = zeromatrix(n,n)
    for i in range(n): 
        for j in range(i,n): 
            a[j][i] = a[i][j] = (i*j+1.)/(i+j+1.) 

    (v,e)= jacobi(a) 

    print ' eigenvalues' 
    printvector(e) 
    #print ' v ' 
    #printmatrix(v) 
    ev = mxm(v,a) 
    for i in range(n):
        norm = dot(v[i],v[i])
        if abs(norm-1.0) > 1e-14:
            print ' Error in eigenvector norm', i, norm
        etest = dot(v[i],ev[i])
        if abs(etest-e[i]) > 1e-14*max(abs(e[0]),abs(e[-1])):
            print ' Error in eigenvalue ', i, e[i], etest
        err = 0.0
        for j in range(n): 
            err = max(err,abs(ev[i][j] - e[i]*v[i][j])) 
        err = err/(n*max(1.0,abs(e[i]))) 
        if err > 1e-12: 
            print ' Error in eigenvector ', i, err
            
    sys.exit()

    print '\n\n   TESTING QUASI-NR SOLVER \n\n' 
    quasinr(f, [1.,0.5,0.3,-0.4], 1e-4, 1e-10) 

    print '\n\n   TESTING GC WITH NUM. GRAD. AND DIAG. PRECOND.\n\n' 
    cgmin2(f, [1.,0.5,0.3,-0.4], 1e-4, 1e-10, reset=20) 

    print '\n\n   TESTING GC WITH ANAL. GRAD. AND WITHOUT OPTIONAL PRECOND.\n\n' 
    cgmin(f, df, [1.,0.5,0.3,-0.4], 1e-4) 

    print '\n\n   TESTING GC WITH ANAL. GRAD. AND WITH OPTIONAL PRECOND.\n\n' 
    cgmin(f, df, [1.,0.5,0.3,-0.4], 1e-4, precond=precond) 

    print '\n\n   TESTING GC WITH ANAL. GRAD. AND NO PRECOND.\n\n' 
    cgminold(f, df, [1.,0.5,0.3,-0.4], 1e-4) 

    print '\n\n    TESTING THE CHOLESKI FACTORIZATION\n\n'
    n = 25
    A = zeromatrix(n,n)
    b = zerovector(n)
    for i in range(n):
        A[i][i] = 10.*(i+1)     # Construct A to be positive definite
        b[i] = 1.0/(i+1.0)
        for j in range(i+1,n):
            A[j][i] = A[i][j] = (i+j)/(2.0+n)
    x = choleski_solve(A,b)
    Ax = mxv(A,x)
    err = 0.0
    for i in range(n):
        err = err + abs(Ax[i] - b[i])
    print ' Cholesky linear equation solution error is ', err



import sys
import collections
import csv

def parsemad(filename):
    f = open(filename,'r')

    energies = []
    conviters = []
    timers = collections.defaultdict(float)
    nnodes = -1
    nthreads = -1
    time = -1
    iter = -1
    energy = -1
    nucrep = -1
    nalpha = nbeta = -1
    norbalpha = norbbeta = -1
    spin_restricted = ''
    aevals = []
    bevals = []
    dipole = -1
    grad = []

    while 1:
        line = f.readline()
        if line=='': break
        data = line.strip().split()

        if data == []:
            pass
        elif data[0] == 'number' and data[1] == 'of' and data[2] == 'electrons':
            nalpha = int(data[3])
            nbeta = int(data[4])
        elif data[0] == 'number' and data[1] == 'of' and data[2] == 'orbitals':
            norbalpha = int(data[3])
            norbbeta = int(data[4])
        elif data[0] == 'Total' and data[1] == 'Dipole' and data[2] == 'Moment:':
            dipole = float(data[3])
        elif data[0] == 'spin' and data[1] == 'restricted':
            spin_restricted = (data[2] == 'true')
        elif data[0] == 'Iteration':
            iter = int(data[1])
        elif data[0] == 'Converged!':
            conviters.append(iter)
            energies.append(energy)
        elif line[0:22] == '                total ':
            energy = float(data[1])
        elif data[0] == 'nuclear-repulsion':
            nucrep = float(data[1])
        elif data[0] == 'alpha' and data[1] == 'eigenvalues':
            data = map(float,f.readline()[3:].strip().split())
            aevals = data
        elif data[0] == 'beta' and data[1] == 'eigenvalues':
            data = map(float,f.readline()[3:].strip().split())
            bevals = data
        elif data[0] == '#nodes':
            nnodes = int(data[1])
        elif data[0] == '#total' and data[1] == 'threads':
            nthreads = int(data[2])
        elif line[0:24] == '         Total wall time':
            time = float(data[3][0:-2])
        elif data[0] == 'timer:':
            t = float(data[-1][0:-1])
            what = data[1]
            for s in data[2:-2]:
                what = what + ' ' + s
            timers[what] += t
        elif line[0:19] == ' Derivatives (a.u.)':
            grad = []
            f.readline(); # skip -----
            f.readline(); # skip blank
            f.readline(); # skip headers
            f.readline(); # skip -----
            while 1:
                line = f.readline()
                data = line.strip().split()
                if len(data) != 7: break
                grad.append(map(float,[data[4], data[5], data[6]]))

    # print "na, nb", nalpha, nbeta
    # print "norba, norbb", norbalpha, norbbeta
    # print "spin restricted", spin_restricted
    # print "nucrep", nucrep
    # print "energies", energies
    # print "iterations", conviters
    # print "nodes", nnodes
    # print "threads", nthreads
    # print "wall time", time
    # print "dipole", dipole
    # print "gradients", grad
    # print "alpha evals", aevals
    # if not spin_restricted:
    #     print " beta evals", bevals
    # print timers

    results = {}
    results["nalpha"] = nalpha
    results["nbeta"] = nbeta
    results["norba"] = norbalpha
    results["norbb"] = norbbeta
    results["spin restricted"] = spin_restricted
    results["nucrep"] = nucrep
    results["energies"] = energies
    results["iterations"] = conviters
    results["nodes"] = nnodes
    results["threads"] = nthreads
    results["wall time"] = time
    results["dipole"] = dipole
    results["gradients"] = grad
    results["alpha evals"] = aevals
    if not spin_restricted:
        results[" beta evals"] = bevals
    results["timers"] = timers

    return results

results = {}
for filename in sys.argv[1:]:
   results[filename] = parsemad(filename)

outf = open('out.csv', 'w')
out = csv.writer(outf)
F0 = results.keys()[0]
for tag in results[F0]['timers']:
    for filename in results.keys():
        nt = results[filename]['threads']
        TBB = (filename[0] == "T")
        tt = "Old"
        if TBB:
            nt = nt - 1
            tt = "TBB"
        out.writerow([tt,tag,nt,results[filename]['timers'][tag]])

for filename in results.keys():
    nt = results[filename]['threads']
    TBB = (filename[0] == "T")
    tt = "Old"
    if TBB:
        nt = nt - 1
        tt = "TBB"
    out.writerow([tt,'total',nt,results[filename]['wall time']])
    

   



def doit(transpose):
    t = ""
    if transpose: t = "T"
    for i in range(2,28,2):
        if i == 12:
            print "#ifdef X86_64"

        print "MTXM_ENTRY(%smTxm%d)" % (t,i)

        jloop = ".%sJLOOP%d" % (t,i)
        print "%s:" % jloop

        for j in range(0,i/2):
            print "ZERO(C%02.2d)" % j

        kloop = ".%sKLOOP%d" % (t,i)
        print "%s:" % kloop
        print "LOADBKJ"

        for j in range(0,i/2):
            print "ABC(%d,C%02.2d)" % (j*16,j)

        print ("INCA")

        for j in range(0,i/2):
            print "ABC2(%d,C%02.2d)" % (j*16,j)

        print "INCA"
        print "sub     $2, NK"
        print "jnz     %s" % kloop

        print "mov     C, B"
        for j in range(0,i/2):
            if transpose:
                print "STORET(C%02.2d)" % j
            else:
                print "STORE(C%02.2d)" % j

        if transpose:
            print "INCCT"
        else:
            print "INCC"
        print "NEXTJ"
        print "jnz     %s" % jloop
        print "RETURN"
    print "#endif"

doit(0)
doit(1)



import sys
from array import array

__kmax = 60
__asciifile = "coeffs"
__checksum = (1.57015282218922890,1.48602374574943450)

def matrix(n,m): 
    a = range(n) 
    for i in range(n): 
        a[i] = array('d',[0.0]*m) 
    return a 

def writemat(k,file,a):
    for i in range(k):
        for j in range(k):
            if a[i][j] == 0.0:
                file.write("0.0\n")
            else:
                file.write("%.17e\n" % a[i][j])

def readmat(k,file):
    a = matrix(k,k)
    for i in range(k):
        for j in range(k):
            a[i][j] = float(file.readline())

    return a

def writeascii(kmax,filename):
    file = open(filename,'w')
    for k in range(1,kmax+1):
        h0,h1,g0,g1 = twoscalecoeffs(k)
        writemat(k,file,h0)
        writemat(k,file,g0)

def readascii(kmax,filename):
    global __cache
    __cache = {}
    file = open(filename,'r')
    for k in range(1,kmax+1):
        h0 = readmat(k,file)
        g0 = readmat(k,file)
        h1 = matrix(k,k)
        g1 = matrix(k,k)
        for i in range(k):
            for j in range(k):
                hphase = (-1)**(i+j)
                gphase = (-1)**(i+j+k)
                h1[i][j] = h0[i][j]*hphase
                g1[i][j] = g0[i][j]*gphase

        __cache[k] = h0,h1,g0,g1

    sum0, sum1 = checksum(kmax)
    if abs(__checksum[0]-sum0) > 1e-16 or \
       abs(__checksum[1]-sum1) > 1e-16:
        print "%.17e %.17e" % __checksum
        print "%.17e %.17e" % (sum0, sum1)
        raise ValueError,"two-scale checksum?"
    
def checksum(kmax):
    sum0 = sum1 = 0.0
    for k in range(1,kmax+1):
        h0,h1,g0,g1 = __cache[k]
        for i in range(k):
            for j in range(k):
                ij = float(j+k)/(i+k)
                sum0 = sum0 + ij*(abs(h0[i][j]) + abs(h1[i][j]) +
                                  abs(g0[i][j]) + abs(g1[i][j]))
                sum1 = sum1 + ij*(h0[i][j] + h1[i][j] +
                                  g0[i][j] + g1[i][j])
                while sum0 > 2.0: sum0 = sum0 * 0.25
                while abs(sum1) > 2.0: sum1 = sum1 * 0.25

    return (sum0, sum1)

if __name__ == "__main__":
    from twoscalecoeffs import twoscalecoeffs
    writeascii(__kmax,__asciifile)
    readascii(__kmax,__asciifile)



#!/usr/bin/python
import sys
if(len(sys.argv) != 2):
    sys.exit("Requires nMAX as an argument");
nMAX = sys.argv[1]
for l in range(0, int(nMAX)):
    for n in range(l+1,int(sys.argv[1])+1):
        print n, l, "0"



#!/usr/bin/env python3
"""Generate vectorized mtxm implementations."""

import logging
import argparse
import sys
from itertools import permutations, product

from codegen.driver import tester_gen
from codegen.bgqdriver import tester_gen_bgq
from codegen.mtxm import *

# TODO: call bgq version of tester gen if needed

def tester(m, versions):
    """Generate .cc files and Makefile to test correctness and
    performance of versions"""
    n = 1
    with open("Makefile", 'w') as mf:
        if m.have_bgp:
            print("\tHPM=/soft/apps/UPC/lib/libhpm.a", file=mf)
            print("\tCXX=tmpixlcxx_r", file=mf)
            print("\tCXXFLAGS=-g -O3 -qarch=450d -qtune=450 -qthreaded", file=mf)
        elif m.have_bgq:
            print("\tCXX=/bgsys/drivers/ppcfloor/comm/xl/bin/mpixlcxx_r", file=mf)
            print("\tCXXFLAGS=-I/bgsys/drivers/ppcfloor/ -O5 -qhot=level=1 -qsimd=auto -qtune=qp -qarch=qp", file=mf)
        else:
            print("\tCXX=g++", file=mf)
            march = "ssse3"
            extra = ''
            if type(m) == MTXMAVX:
                march = "avx"
            if type(m) == MTXMAVX2:
                march = "avx2"
                extra = "-march=core-avx2"
            print("\tCXXFLAGS=-O3 {} -m{} -lrt -lm".format(extra, march), file=mf)
        print("all:", file=mf)

        while versions:
            bunch_size = 1
            if m.have_bgp:
                bunch_size = 64
            elif m.have_bgq:
                bunch_size = 128
            cur, versions = versions[:bunch_size], versions[bunch_size:]
            with open("tune_mtxm_{}.cc".format(n), 'w') as f:
                if m.have_bgq:
                    tester_gen_bgq(f, cur, m.gen, m.complex_a, m.complex_b)
                else:
                    tester_gen(f, cur, m.gen, m.complex_a, m.complex_b, m.have_bgp)
            print("\t$(CXX) -c $(CXXFLAGS) tune_mtxm_{0}.cc -o tune_mtxm_{0}.o".format(n), file=mf)
            print("\t$(CXX) $(CXXFLAGS) tune_mtxm_{0}.o $(HPM) -o tune_mtxm_{0}.x".format(n), file=mf)
            if m.have_bgq:
                print("\tqsub -A MADNESS_MPQC_esp -t 10 -n {0} --mode c1 --cwd `pwd` -O tune_mtxm_{1} ./tune_mtxm_{1}.x".format(len(cur), n), file=mf)
            n += 1

def main():
    # Parse Arguments
    parser = argparse.ArgumentParser(description='What does this program do?')
    parser.add_argument('-l', '--log', dest='loglevel', default='ERROR',
            help='log level to output (default: ERROR)')
    parser.add_argument('--log-file', dest='logfile',
            help='destination file to write logging output')
    parser.add_argument('-a', '--complex-a', action='store_true', default=False,
            help='A (left) matrix is complex', dest='cxa')
    parser.add_argument('-b', '--complex-b', action='store_true', default=False,
            help='B (right) matrix is complex', dest='cxb')
    parser.add_argument('-m', '--arch', default='sse', choices=['sse', 'avx', 'avx2', 'bgp', 'bgq'],
            help='Target architecture')
    parser.add_argument('-n', '--name', default='mtxmq',
            help='Name of function to generate')
    parser.add_argument('-t', '--test', default=False, action='store_true',
            help='Generate performance and correctness testing code')
    args = parser.parse_args()

    try:
        loglevel = getattr(logging, args.loglevel.upper())
    except:
        parser.error("Invalid LOGLEVEL")

    # Configure logging
    logging.basicConfig(level=loglevel,
            format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
            filename=args.logfile)
    logger = logging.getLogger(__name__)

    if args.arch == 'sse':
        M = MTXMSSE
        bests = {
                (False, False): ("jik", {'i':2, 'j':14, 'k':1}, args.name),
                (True, False) : ("jik", {'i':2, 'j':12, 'k':1}, args.name),
                (False, True) : ("jik", {'i':2, 'j':14, 'k':1}, args.name),
                (True, True)  : ("ijk", {'i':2, 'j':14, 'k':1}, args.name)}
        versions = [(lo, {'i':i, 'j':j, 'k':1}, '{}_i{}j{}k1'.format(lo, i,j)) for i in range(2,21,2) for j in range(2,21,2) for lo in ["ijk", "jik"] if i*j <= 60]
    elif args.arch == 'avx':
        M = MTXMAVX
        bests = {
                (False, False): ("jik", {'i':2, 'j':20, 'k':1}, args.name), # jik_i3j12k1 is a bit better, except on 400 20 20 much slower
                (True, False) : ("ijk", {'i':4, 'j':12, 'k':1}, args.name), # ijk_i4j8k1 is better for 400 20 20 but worse on others
                (False, True) : ("ijk", {'i':3, 'j':12, 'k':1}, args.name), # jik_i3j12k1 best on avg across tested sizes, ijk_i4j8k1 best on 400 20 20
                (True, True)  : ("ijk", {'i':2, 'j':16, 'k':1}, args.name)} # jik is better for squares, ijk better for 400 20 20 and similar
        versions = [(lo, {'i':i, 'j':j, 'k':1}, '{}_i{}j{}k1'.format(lo, i,j)) for i in range(1,10) for j in range(4,21,4) for lo in ["ijk", "jik"] if i*j <= 60]
    elif args.arch == 'avx2':
        M = MTXMAVX2
        bests = {
                (False, False): ("jik", {'i':2, 'j':20, 'k':1}, args.name),
                (True, False) : ("jik", {'i':3, 'j':16, 'k':1}, args.name),
                (False, True) : ("jik", {'i':3, 'j':16, 'k':1}, args.name),
                (True, True)  : ("ijk", {'i':2, 'j':20, 'k':1}, args.name)}
        versions = [(lo, {'i':i, 'j':j, 'k':1}, '{}_i{}j{}k1'.format(lo, i,j)) for i in range(1,10) for j in range(4,21,4) for lo in ["ijk", "jik"] if i*j <= 60]
    elif args.arch == 'bgp':
        M = MTXMBGP
        bests = {
                (False, False): ("ijk", {'i':8, 'j':4, 'k':1}, args.name),
                (True, False) : ("ijk", {'i':6, 'j':8, 'k':1}, args.name),
                (False, True) : ("jik", {'i':2, 'j':20, 'k':1}, args.name), # ijk_i9j4k1 is better for (m*m,m)T*(m*m) 400 20 20 by almost 30%, but generally quite a bit worse
                (True, True)  : ("ijk", {'i':5, 'j':8, 'k':1}, args.name)}
        versions = [(lo, {'i':i, 'j':j, 'k':1}, '{}_i{}j{}k1'.format(lo, i,j)) for i in range(1,10) for j in range(4,21,4) for lo in ["ijk", "jik"] if i*j <= 60]
    elif args.arch == 'bgq':
        M = MTXMBGQ
        bests = {
                (False, False): ("jik", {'i':4, 'j':20, 'k':1}, args.name),
                (True, False) : ("ijk", {'i':6, 'j':8, 'k':1}, args.name),
                (False, True) : ("jik", {'i':4, 'j':20, 'k':1}, args.name),
                (True, True)  : ("jik", {'i':3, 'j':28, 'k':1}, args.name)}
        versions = [(lo, {'i':i, 'j':j, 'k':1}, '{}_i{}j{}k1'.format(lo, i,j)) for i in range(1,14) for j in range(4,29,4) for lo in ["ijk", "jik"] if i*j <= 90]

    m = M(args.cxa, args.cxb)
    best = bests[(args.cxa, args.cxb)]

    if args.test:
        tester(m, versions)
    else:
        m.gen(sys.stdout, *best)


if __name__ == '__main__':
    main()




from math import *
from types import *
import re

# Outstanding issues 
#   Convergence criteria for series expansion of functions
#   How to round correctly?
#   Also, rounding while printing
#   Faster truncation
#   Use of interp routines for function evaluation

class interptable:
    def __init__(self, func, lo, hi, n, precomputed=None):
        '''
        Tablulate the function over n intervals or n+1 points
        in [lo,hi] inclusive of the end points.

        This is now hardwired to longfloat, but the changes
        to move it back to generality are obvious.
        '''
        if precomputed:
            (lo, hi, n, h, h2, scale, func, x, f, nbits) = precomputed
        else:
            lo = longfloat(lo)
            hi = longfloat(hi)
            h = (hi - lo)/(n+1)
            h2 = h*0.5
            scale = 1.0/h
            x = [0]*(n+1)
            f = [0]*(n+1)
            for i in range(n+1):
                x[i] = lo + i*h
                f[i] = func(x[i])
            nbits = longfloat.nbits
        self.lo = lo
        self.hi = hi
        self.n = n
        self.h = h
        self.h2 = h2
        self.scale = scale
        self.func = func
        self.x = x
        self.f = f
        self.nbits = nbits

    def map(self,x):
        '''
        If x is within the range of the interpolation table return
        k - the index of the nearest node x[k]
        d - the difference x-x[k]

        If x is out of range raise a value exception.
        '''
        if self.lo <= x <= self.hi:
            k = int((x-self.lo)*self.scale)
            d = x - self.x[k]
            if d > self.h2:
                k = k + 1
                d = x - self.x[k]
            return (k,d)
        else:
            raise ValueError, ('when interpolating %s' % self.func)

    def __repr__(self):
        '''
        This so that tables may be precomputed for fast startup.
        '''
        return 'interptable(None,None,None,None,precomputed=' + \
              repr((self.lo,self.hi,self.n,self.h,self.h2,self.scale,
                    str(self.func),self.x,self.f,self.nbits)) + ')'

class interpolators:
    def __init__(self):
        # Exp tabulated on [-1/2,1/2]
        half = longfloat((-1,1))
        # 20 .12
        # 40 .12
        #100 .10
        #400 .088
        ###self.texp = interptable(lambda (x):x.exp(), -half, half, 40)
        #print self.texp

        # Sin and Cos on [0,pi/2] (1.571 is slightly bigger than pi/2)
        zero = longfloat(0)
        pi2  = longfloat(1.571)
        ###self.tsin = interptable(lambda (x):x.sin(), zero, pi2, 30)
        ###self.tcos = interptable(lambda (x):x.cos(), zero, pi2, 30)

        # log(x) with x in [1/sqrt(2), sqrt(2)]
        # 1.415 > sqrt(2)    0.707 < 1/sqrt(2)
        n = 100
        self.tlog = interptable(lambda (x):x.log(),
                                longfloat(0.707), longfloat(1.415), n)
        self.tlog.rx = [0]*(n+1)
        for k in range(n+1):
            self.tlog.rx[k] = self.tlog.x[k].reciprocal()

    def exp(self,x):
        '''
        For arguments in the range [0,1/2] exp(x) is O(1) so it
        suffices to ensure that the remainder is < eps
        '''
        (k, d) = self.texp.map(x)
        #print k , d
        fk = self.texp.f[k]
        result = fk
        term = fk
        eps = x.eps()
        for i in range(1,50):
            term = term * d / i
            result = result + term
            if abs(term) < eps:
                #print ' exp terminating at ', i
                return result
        raise ArithmeticError, "interp exp did not converge"

    def sin(self,x):
        '''
        For arguments in the range [0,pi/2] sin(x) is O(min(x,1)).
        So, unless x is a hard zero, use relative error.
        '''
        if not x:
            return longfloat(0)
        (k, d) = self.tsin.map(x)
        sink = self.tsin.f[k]
        cosk = self.tcos.f[k]
        result = sink
        term = d
        eps = x.eps()
        if x < 1:
            eps = eps*x
        for i in range(1,50,4):
            result = result + cosk*term
            term = term*d/(i+1)
            result = result - sink*term
            term = term*d/(i+2)
            result = result - cosk*term
            term = term*d/(i+3)
            result = result + sink*term
            term = term*d/(i+4)
            if abs(term) < eps:
                #print ' sin terminating at ', i+3
                return result
        raise ArithmeticError, "interp sin did not converge"
        
    def cos(self,x):
        '''
        Like sin, use relative error.
        '''
        if not x:
            return longfloat(1)
        (k, d) = self.tcos.map(x)
        sink = self.tsin.f[k]
        cosk = self.tcos.f[k]
        result = cosk
        term = d
        eps = x.eps()
        if x < 1:
            eps = eps*x
        for i in range(1,50,4):
            result = result - sink*term
            term = term*d/(i+1)
            result = result - cosk*term
            term = term*d/(i+2)
            result = result + sink*term
            term = term*d/(i+3)
            result = result + cosk*term
            term = term*d/(i+4)
            if abs(term) < eps:
                #print ' cos terminating at ', i+3
                return result
        raise ArithmeticError, "interp sin did not converge"

    def log(self, x):
        '''
        log(x+d) = log(x) - sum(i=1..inf) (-d/x)^i/i

        For x in the range [1/root(2),root(2)] log(x) is about
        O(x-1) ... so use relative error to this for convergence.

        log(x) = 2*(p + p^3/3 + p^5/5 + ...) where p = (x-1)/(x+1)

        log(xk+d) = log(xk)+log(1+d/xk)

        '''
        if x == 1:
            return longfloat(0)
        (k, d) = self.tlog.map(x)
        dx = d*self.tlog.rx[k]
#        print k, float(d), float(self.tlog.x[k]), float(dx)
        p = dx/(dx+longfloat(2))
#        print ' p', float(p)
        p2 = p*p
        eps = x.eps() * abs(self.tlog.f[k])
#        print ' eps ', float(eps)
        term = longfloat(1)
        result = longfloat(0)
        for i in range(1,501,2):
            result = result + term/longfloat(i)
            term = term*p2
            if abs(term) < eps:
                break
        if i >= 498:
            raise ArithmeticError, "interp log did not converge"
        return (2*p*result + self.tlog.f[k])
        
##         dx = d/x
##         result = self.tlog.f[k]
##         term = dx
##         eps = x.eps() * abs(x-1) # relative error
##         for i in range(1,50):
##             result = result + term/i
##             term = term * dx
##             if abs(term) < eps:
##                 #print ' log terminating at ', i
##                 return result
##         raise ArithmeticError, "interp log did not converge"

    def sindsin(self, x):
        return (self.sin(x), self.cos(x))

    def cosdcos(self, x):
        return (self.cos(x), -self.sin(x))

    def expdexp(self, x):
        return (self.exp(x), self.exp(x))

    def logdlog(self, x):
        return (self.log(x), 1/x)

    def invert(self, y, gdg, x):
        '''
        solve for x in y = g(x) ... i.e., invert g assuming that
        we can compute both g and its derivative.

        x inputs an initial guess for x which had better be in the
        quadratically convergent region.
        '''
        # delta = (g(x) - y)/g'(x)
        eps = x.eps()
        while 1:
            (g,dg) = gdg(x)
            print ' g dg ', g, dg
            delta = (g - y)/dg
            print ' delta ', x, delta
            x = x - delta
            if abs(delta) < eps:
                break
        return x

class longfloat:
    '''
    Represents a floating point number to arbitrary precision
    by storing the mantissa as a long integer and the exponent
    as an integer.

    value = 2**exponent * mantissa

    The class member nbits is used to control the number of bits
    of precision retained by operations.  It defaults to 208 which
    is four times that of the standard double precision (52 bits).
    It may be modified as follows
    
    longfloat.nbits = newvalue
    
    and the modified value will be adopted by all new instances.
    
    The class defines the standard methods to support
    -  conversion to/from integers, floats, longs, strings, 2-tuple
    -  addition, subtraction, multiplication, negation, division
    -  comparison and non-zero test
    -  printing and representation as a string (__repr__)

    and also defines methods to compute the following
    
    -  reciprocal, reciprocal square root and square root

    After each operation the result is truncated with rounding to
    the required finite precision.  Internally, some operations
    will be performed to a higher precision.

    The method eps() returns the smallest number such that 1+eps > 1.

    Some rather limited testing suggests that the errors are
    -  the last bit in the basic operations +*- due to rounding
    -  the last two bits in divide and square root
    -  the last two or three bits in log/exp/sin/cos
    -  the last four or five bits in asin
    -  the last five to eight bits in acos
    '''

    nbits   = 208
    debug   = 0

    __name__ = 'longfloat'

    __small_int_recip = [(0,0)]*501  # Cache 1/n  for 0<n<=500 ... entry is (1/n,nbits)
   
    def __init__(self, value, nbits=None):
        # Constructor allows override of class default for nbits
        if nbits != None:
            self.nbits = nbits
        if self.nbits <= 0:
            raise ValueError, "longfloat.nbits must be > 0"
            
        # Initialization from ints and longfloat done here, others done
        # thru coerce
        t = type(value)
        if isinstance(value,longfloat):
            self.exponent = value.exponent
            self.mantissa = value.mantissa
        elif t == IntType or t == LongType:
            self.exponent = 0
            self.mantissa = long(value)
        else:
            (junk, a) = self.__coerce__(value)
            self.exponent = a.exponent
            self.mantissa = a.mantissa

    def __power2(self, exponent):
        ' return 2.0**exponent '
        if exponent < 0:
            fac = 0.5
        else:
            fac = 2.0
        
        scale = 1.0
        for i in range(abs(exponent)):
            scale = scale * fac
        return scale
    
    def __coerce__(self, other):
        ' coerce strings, ints, longs, floats and longfloats into longfloat'
        if isinstance(other,longfloat):
            return (self, other)

        t = type(other)
        if self.debug: print ' longfloat: coercing ', t, other

        value = longfloat(0)
        if (t == IntType) or (t == LongType):
            value.exponent = 0
            value.mantissa = long(other)
        elif t == StringType:
            # The exact representation from __repr__() is a tuple of 
            # two integers (exponent, mantissa).
            # Alternatively match a floating point number
            # Match two signed integers from a 2-tuple, e.g., (1,2L)
            pat = r'\([ ]*([+-]?\d+)[lL]?[ ]*,[ ]*([+-]?\d+)[lL]?[ ]*\)'
            m = re.search(pat,other)
            if m:
                (e,m) = m.groups()
                value.exponent = int(e)
                value.mantissa = long(m)
            else:
                patf = r'\A\s*([+-]?\d+){1,1}(\.\d*)?([eE][+-]?\d+)?'
                m = re.search(patf,other)
                mantissa = ''
                if m:
                    mantissa = ''
                    e10 = 0
                    (before,after,expnt) = m.groups()
                    print ' before, after, expnt ', before, after, expnt
                    if before:
                        mantissa = mantissa + before
                    if after:
                        after = after[1:]        # Get rid of leading .
                        e10 = -len(after)
                        mantissa = mantissa + after
                    if expnt:
                        e10 = e10 + int(expnt[1:])
                    print ' mantissa ', mantissa, ' e10 ', e10
                    if e10:
                        longfloat.nbits = longfloat.nbits + 31
                        log10  = self.__log10()
                        log2   = self.__log2()
                        log210 = log10/log2
                        e2 = e10 * log210        # Base two exponent
                        r2 = e2 - int(e2)        # Fractional part
                        print ' r2 ', float(r2)
                        e2 = int(e2)             # Integer part
                        r2 = (r2 * log2).exp()   # 2^r2
                        print ' 2^r2 ', float(r2)
                        print ' r2* ', float(r2 - r2.log()/log2)
                        value.exponent = e2
                        value.mantissa = long(mantissa)
                        value = value * r2
                        longfloat.nbits = longfloat.nbits - 31
                        value.truncate()
                    else:
                        value.exponent = 0
                        value.mantissa = m
                else:
                    raise ("longfloat: fromstr: '%s' is neithr a tuple (exponent,mantissa) nor a floating point number" % other)
        elif (t == TupleType) and (len(other) == 2):
            value.exponent = int(other[0])
            value.mantissa = long(other[1])
        elif t == FloatType:
            # A double has a 52 bit mantissa and we need to preserve
            # all of those bits ... but storing too many is inefficient.
            # Must also watch out for very small numbers that will
            # overflow when reciprocated

            # This routine is paranoid and always checks that an exact
            # representation of the input was generated.

            if other == 0.0:
                exponent = 0
                mantissa = 0L
            else:
                exponent = int(log(abs(other))/log(2.0))
                exponent = exponent - 53
                extra = 0
                while exponent < -990:
                    other = other * 2.0
                    exponent = exponent + 1
                    extra = extra + 1
                mantissa = long(other*self.__power2(-exponent))
                exponent = exponent - extra
                while extra > 0:
                    other = other * 0.5
                    extra = extra - 1

                # Eliminate trailing zeros in the mantissa for
                # values that are actually exactly representable
                while (mantissa & 1L) == 0:
                    mantissa = mantissa >> 1
                    exponent = exponent + 1
                
            test = float(mantissa)*self.__power2(exponent)
            if (test - other) != 0.0:
                print ' exponent ', exponent
                print ' mantissa ', mantissa
                print ' input      %24.17e' % other
                print ' test       %24.17e' % test
                print ' err        %24.17e' % (test-other)
                raise TypeError, "failed to convert double to longfloat"
                
            value.mantissa = mantissa
            value.exponent = exponent
        else:
            print ' unknown TYPE IS ', t
            raise TypeError, 'conversion to longfloat failed'

        return (self,value)

    def __float__(self):
        # If the mantissa is too long then the standard conversion
        # of long to float does not work.  Ugh.

        e = self.exponent
        m = self.mantissa
        while abs(m) > 1e17:
            m = m >> 1
            e = e + 1
        return (float(m)*self.__power2(e))  # THIS CAN FAIL .. FIX IT!!!!

    def __int__(self):
        return int(long(self))

    def __long__(self):
        e = self.exponent
        m = self.mantissa
        if e > 0:
            m = m << e
        else:
            m = m >> abs(e)
        return m
    
    def __repr__(self):
        ' exact string representation is a tuple (exponent,mantissa) '
        return "longfloat((" + repr(self.exponent) + "," + repr(self.mantissa) + "))"

    def __str__(self):
        ' human readable but inexact string representation is decimal float'
        if not self:
            return '0'
            
        # Need to get the exponent into base 10 and scale the
        # mantissa by its fractional part.
        # 2^e * m = 10^(e/log2(10)) * m = 10^e10 * (10^r10 * m)
        # base 10 exponent is e10+r10 (integer part + remainder)

        nbits   = self.nbits
        ndigits = int((self.nbits-1)*0.30103) + 1
        longfloat.nbits = longfloat.nbits + 31 # Should be log2(|e+nbm|)?

        x = longfloat(self)
        e = x.exponent
        m = x.mantissa
        nbm = x.__nbits_in_mantissa()
        if nbm < nbits:
            e = e - (nbits - nbm)
            m = m << (nbits - nbm)
            x.exponent = e
            x.mantissa = m

        log10  = self.__log10()
        log2   = self.__log2()
        log210 = log10/log2

        # Split exponent into fractional and integer parts
        e10  = e / log210
        r10 = e10 - int(e10)
        e10 = int(e10)
        tenr10 = (r10*log10).exp()
        
        # Generate 10^r10 * m truncated to the appropriate no. of bits
        tmp = longfloat(x)
        tmp.exponent = 0
        tmp = tmp*tenr10

        # Shift the mantissa to zero the remaining exponent
        etmp = tmp.exponent
        mtmp = tmp.mantissa
        m10 = tmp.mantissa

        if etmp > 0:
            m10 = m10 << etmp
        elif etmp < 0:
            # Cannot just shift ... must round correctly.
            # Examine the last bit discarded
            etmp = abs(etmp)
            m10 = m10 >> (etmp-1)
            if (m10 & 1):
                m10 = m10 + 2
            m10 = m10 >> 1
            
        # Now we have e10 and m10
        # Count the no. of decimal figures in m10 so that we can
        # have one significant figure before the decimal point
        strm = str(abs(m10))
        n10 = len(strm) - 2
        e10 = e10 + n10

        # Finally can generate the result
        signm = '+'
        if m10 < 0: signm = '-'
        signe = '+'
        if e10 < 0: signe = '-'

        lead = strm[0]
        strm = strm[1:]
        if strm[-1] == 'L':
            strm = strm[:-1]
        while len(strm) and (strm[-1] == '0'):
            strm = strm[:-1]
        if strm == '':
            strm = '0'
        
        longfloat.nbits = longfloat.nbits - 31

        return signm + lead + '.' + strm + 'e' + signe + str(abs(e10))
        

    def __mul__(self,other):
        value = longfloat(0)
        value.mantissa = self.mantissa*other.mantissa
        value.exponent = self.exponent+other.exponent
        value.truncate()
        return value

    def __rmul__(self,other):
        return self*other

    def __common_exponent(self,other):
        '''
        Return the mantissas scaled to a common exponent
        '''
        eself = self.exponent
        mself = self.mantissa
        eother = other.exponent
        mother = other.mantissa

        if eother < eself:
            exponent = eother
            mself = mself << (eself-eother)
        else:
            exponent = eself
            mother = mother << (eother-eself)
        return (exponent, mself, mother)

    def __add__(self,other):
        (exponent, mself, mother) = self.__common_exponent(other)
        value = longfloat(0)
        value.exponent = exponent
        value.mantissa = mself + mother
        value.truncate()
        return value
            
    def __radd__(self,other):
        return self+other

    def __sub__(self,other):
        return self+(-other)

    def __rsub__(self,other):
        return other+(-self)

    def __neg__(self):
        value = longfloat(0)
        value.exponent =  self.exponent
        value.mantissa = -self.mantissa
        return value

    def __abs__(self):
        value = longfloat(0)
        value.exponent =  self.exponent
        value.mantissa = abs(self.mantissa)
        return value
        
    def __pos__(self):
        return self

    def __nonzero__(self):
        return (self.mantissa != 0L)

    def __cmp__(self, other):
        (exponent, mself, mother) = self.__common_exponent(other)
        if mself < mother:
            return -1
        elif mself == mother:
            return 0
        else:
            return 1

    def __pow__(self,power):
        if power == 0:
            return longfloat(1)
        elif power == 1:
            return longfloat(self)
        elif power == 2:
            return self*self
        elif power == 3:
            return self*self*self
        elif power == 4:
            x2 = self*self
            return x2*x2

        if power < 0 and self == 0:
            raise ValueError, "longfloat: pow: negative power of zero"

        x = longfloat(self)
        phase = 1
        if power.exponent == 0:
            # Integer power ... can handle negative arguments
            if x < 0:
                if (power % 2):
                    phase = -1
                x = -x
            # Repeated squaring for integer powers
            m = int(power.mantissa)
            if m < 0:
                x = x.reciprocal()
                m = -m
            xk = x                  # x^(2^k)
            result = 1.0
            while m:
                if (m&1):
                    result = result * xk
                xk = xk*xk
                m = m >> 1
            if phase < 0:
                result = -result
            return result
        
        if x < 0:
            raise ValueError, "longfloat: pow: negative number to a non-integer power"

        logx = x.log()
        result = (power*logx).exp()
        if phase < 0:
            result = -result

        return result

    def __rpow__(self,power):
        return power.__pow__(self)

    def rsqrt(self):
        '''
        Compute the reciprocal squareroot using Newton-Raphson iteration.
        '''
        if self <= 0.0:
            raise ValueError, 'reciprocal squareroot of non-positive value'

        nbits = self.nbits

        # Generate the initial guess using the standard float operation
        # but we need to make sure it is representable.
        nbm = self.__nbits_in_mantissa()
        e = self.exponent
        if ((e+nbm)%2):
            nbm = nbm + 1
        self.exponent = -nbm
        s = longfloat(1.0/sqrt(float(self)))
        self.exponent = e
        s.exponent = s.exponent - ((e + nbm)>>1)

        threehalf = longfloat(3)
        threehalf.exponent = threehalf.exponent - 1
        xhalf = longfloat(self)
        xhalf.exponent = xhalf.exponent - 1
        nbitsgot = 51
        while nbitsgot < nbits:
            #snew = half*s*(three - s*s*self)
            snew = s*(threehalf - s*s*xhalf)
            s = snew
            nbitsgot = nbitsgot*2-1 # Quad. conv. with rounding loss

        return s

    def sqrt(self):
        if self.mantissa == 0L:
            return longfloat(0)
        else:
            return self*self.rsqrt()

    def reciprocal(self):
        '''
        Compute the reciprocal using Newton-Raphson iteration.
        '''
        if not self:
            raise ValueError, 'reciprocal of zero value'

        if not self.exponent:
            # We have an integer.  The reciprocals of small integers
            # are precomputed and cached for efficiency
            r = self.__small_integer_reciprocal()
            if r:
                return r
        
        nbits = self.nbits

        # Generate the initial guess using the standard float operation
        # but we need to make sure it is representable.
        nbm = self.__nbits_in_mantissa()
        e = self.exponent
        self.exponent = -nbm
        r = longfloat(1.0/float(self))
        self.exponent = e
        r.exponent = r.exponent - (e + nbm)
        
        two = longfloat(2)
        nbitsgot = 51
        while nbitsgot < nbits:
            rnew = r*(two - r*self)
            r = rnew
            nbitsgot = nbitsgot*2 - 1 # Quad. conv. with rounding loss

        return r

    def __small_integer_reciprocal(self):
        ''' Maintain a cache of the reciprocals of small integers '''
        
        m = self.mantissa
        absm = abs(m)
        if absm > 500:
            return None
        absm = int(absm)
        
        (val,nbits) = longfloat.__small_int_recip[absm]
        if nbits < longfloat.nbits:
            val = 0
            
        if not val:
            #print ' computing small int ', absm
            # Not in cache. Set expnt = -1 to stop reciprocal calling
            # here in infinite loop.
            val = longfloat((-1,absm)).reciprocal()
            val.exponent = val.exponent - 1
            longfloat.__small_int_recip[absm] = (val,longfloat.nbits)
        
        # Finally, return a copy of the cached object
        val = longfloat(val).truncate()
        if m < 0:
            return -val
        else:
            return val
        
    def __log10(self):
        # Cache the value of log10 in the class but we
        # need to keep track of the precision used to compute it.
        try:
            if longfloat.__log10nbits >= self.nbits:
                return longfloat(longfloat.__log10value)
        except AttributeError: 
            pass

        longfloat.__log10value = longfloat(10).log()
        longfloat.__log10nbits = self.nbits
        return longfloat(longfloat.__log10value)

    def __log2(self):
        # Cache the value of log2 in the class but we
        # need to keep track of the precision used to compute it.
        try:
            if longfloat.__log2nbits >= self.nbits:
                return longfloat(longfloat.__log2value)
        except AttributeError: 
            pass
        
        longfloat.nbits = longfloat.nbits + 8

        test = longfloat((-self.nbits-4,1))
        x = longfloat(2)
        p = (x-1)/(x+1)
        p2 = p*p
        term = longfloat(1)
        result = longfloat(0)
        i = 1
        while 1:
            result = result + term/i
            term = term*p2
            i = i + 2
            if abs(term) < test:
                break
        log2 = longfloat.__log2value = 2*p*result
        longfloat.__log2nbits = self.nbits

        longfloat.nbits = longfloat.nbits - 8

        return longfloat(log2)

    def log(self):
        '''
        Return the natural log of self. self is not modified.

        Use log(x) = log(x * 2^-k * 2^k) = k*log(2) + log(x*2^-k)
        to scale x into (1/2) <= x <= 1.

        Then for the scaled variable use

        log(x) = 2*(p + p^3/3 + p^5/5 + ...) where p = (x-1)/(x+1)
        '''

        # CANNOT CALL PRINT OF LONGFLOAT FROM THIS ROUTINE SINCE PRINT
        # CALLS THIS !!!!!
        
        if self <= 0:
            raise ValueError, "longfloat: taking log of non-positive value"

        longfloat.nbits = longfloat.nbits + 8

        e = self.exponent
        m = self.mantissa

        # Initially scale the argument into [0.5,1], then into
        # [1/sqrt(2), sqrt(2)] which gives the smallest range for p.
        k = self.__nbits_in_mantissa()
        power = e + k

        x = longfloat((-k,m))
        x707 = longfloat(181)
        x707.exponent = -8    # 181/256 = 0.707
        if abs(x) < x707:
            x.exponent = x.exponent + 1
            power = power - 1
            
        p = (x-1)/(x+1)
       
        # Apply the standard power-series
        test = longfloat((-self.nbits-4,1))*abs(x-1)
        term = longfloat(1)
        p2 = p*p
        result = longfloat(0)
        for i in range(1,501,2):
            result = result + term/i
            term = term*p2
            if abs(term) < test:
                break
        if i >= 498:
            raise ArithmeticError, "longfloat log did not converge"

        result = 2*p*result

        longfloat.nbits = longfloat.nbits - 8

        result = power*self.__log2() + result
        result.truncate()
        return result

    def exp(self):
        '''
        Return exp(self).  self is not modified.
        '''
        # CANNOT CALL PRINT OF LONGFLOAT FROM THIS ROUTINE SINCE PRINT
        # CALLS THIS !!!!!

        if not self:
            return longfloat(1)
        
        longfloat.nbits = longfloat.nbits + 8

        e = self.exponent
        m = self.mantissa

        # Scale the argument so that it is less than 1/2 (try 1/4)
        k = self.__nbits_in_mantissa()
        k = k + 1                       # to get 1/4 rather than 1/2
        power = e + k
        if power < 0:
            power = 0

        x = longfloat(0)
        x.exponent = e - power
        x.mantissa = m
            
        # Apply the standard power-series
        test = longfloat((-self.nbits-4,1))*abs(x)
        term = longfloat(1)
        result = longfloat(0)
        for i in range(1,500):
            result = result + term
            term = term * x / i
            if abs(term) < test:
                break
        if i >= 499:
            raise ArithmeticError, "longfloat exp did not converge"
        
        while power > 0:
            result = result * result
            power = power - 1

        longfloat.nbits = longfloat.nbits - 8
        result.truncate()

        return result

    def pi(self):
        '''
        Compute pi with the quadratically convergent Borwein algorithm
        which has monotonically decreasing error pin-pi < 10^-2^(n+1)
        '''
        try:
            if longfloat.__pinbits >= self.nbits:
                p = longfloat(longfloat.__pi)
                p.truncate()
                return p
        except AttributeError: 
            pass
            
        longfloat.nbits = longfloat.nbits + 8
        
        half = longfloat(0.5)
        x = longfloat(2).sqrt()         # x0
        mypi = 2 + x                      # pi0
        y = x.sqrt()                    # y1
        rsqrtx = x.rsqrt()
        sqrtx  = x*rsqrtx
        x = half*(sqrtx + rsqrtx)    # x1

        test = longfloat((-self.nbits/2-1,1))
        for iter in range(30):
            ryn1 = (y+1).reciprocal()
            mypi_new = mypi*(x+1)*ryn1
            err = mypi - mypi_new
            mypi = mypi_new
            if err < test:
                break
            rsqrtx = x.rsqrt()
            sqrtx  = x*rsqrtx
            y = (y*sqrtx + rsqrtx)*ryn1
            x = half*(sqrtx + rsqrtx)    # x1
        if iter >= 29:
            raise ArithmeticError, 'longfloat pi did not converge'

        mypi.nbits = longfloat.nbits = longfloat.nbits - 8
        mypi.truncate()

        longfloat.__pi = mypi
        longfloat.__pinbits = mypi.nbits

        return longfloat(mypi)

    def __sin(self):
        ''' Compute sine of argument already scaled to [0,pi/4] '''

        x  = self
        x2 = x*x
        
        # Apply the standard power-series
        test = longfloat((-self.nbits-4,1))
        term = longfloat(x)
        result = longfloat(0)
        i = 1
        for i in range(1,301,2):
            result = result + term
            term = -term * x2 / (i+1) / (i+2) # Recips of small ints are fast
            if abs(term) < test:
                break
        if i >= 299:
            raise ArithmeticError, "longfloat sin did not converge"
            
        return result

    def __cos(self):
        ''' Compute cosine of argument already scaled to [0,pi/4] '''

        x  = self
        x2 = x*x
        
        # Apply the standard power-series
        test = longfloat((-self.nbits-4,1))
        term = longfloat(1)
        result = longfloat(0)
        for i in range(0,300,2):
            result = result + term
            term = -term * x2 / (i+1) / (i+2)  # Recips of small ints are fast
            if abs(term) < test:
                break
        if i >= 298:
            raise ArithmeticError, "longfloat cos did not converge"
            
        return result

    def sin(self):
        longfloat.nbits = longfloat.nbits + 8
        x = longfloat(self)
        # Shift argument into [0,2pi]
        pi    = self.pi()
        twopi = pi+pi
        pi2   = pi*0.5
        pi4   = pi*0.25
        n     = int(x/twopi)
        x     = x - n*twopi

        # Shift argument into [0,pi/2] using
        # sin(x) = sin(pi-x)=-sin(pi+x)=-sin(2pi-x)
        phase = 1
        if x > pi:
            phase = -1
            x = x - pi
        if x > pi2:
            x = pi - x

        # Shift into [0,pi/4] using
        # sin(x+pi/4) = (sin(x) + cos(x)) / sqrt(2)

        if x >= pi4:
            x = x - pi4
            result = phase * (x.__sin() + x.__cos()) * longfloat(2).rsqrt()
        else:
            result = phase * x.__sin()

        longfloat.nbits = longfloat.nbits - 8

        result.truncate()
        return result

    def cos(self):
        longfloat.nbits = longfloat.nbits + 8
        x = longfloat(self)
        # Shift argument into [0,2pi]
        pi    = self.pi()
        twopi = pi+pi
        pi2   = pi*0.5
        pi4   = pi*0.25
        n     = int(x/twopi)
        x     = x - n*twopi

        # Shift argument into [0,pi/2] using
        # cos(x) = -cos(pi-x) = -cos(pi+x)= cos(2pi-x)
        phase = 1
        if x > pi:
            x = x - pi
        if x > pi2:
            phase = -1
            x = pi - x

        # Shift into [0,pi/4] using
        # cos(x+pi/4) = (cos(x) - sin(x)) / sqrt(2)

        if x >= pi4:
            x = x - pi4
            result = phase * (x.__cos() - x.__sin()) * longfloat(2).rsqrt()
            ##cosx = x.__cos()
            ##sinx = ((1+cosx).sqrt())*((1-cosx).sqrt())
            ##result = phase * (cosx - sinx) * longfloat(2).rsqrt()
        else:
            result = phase * x.__cos()

        longfloat.nbits = longfloat.nbits - 8
        result.truncate()
        return result

    def acos(self):
        '''
        pi/2 - asin(x) -1 < x < 1
        '''
        pi  = self.pi()
        pi2 = 0.5*pi

        # Must first shift the argument into a range where the subtraction
        # pi/2 - asin(x) does not lose precision

        x = longfloat(self)
        xold = longfloat(x)
        # Restrict to 0 <= x <= 1 with
        # acos(x) = pi - acos(-x)
        extra = 0
        phase = 1
        if x < 0:
            extra = pi
            phase = -1
            x = -x

        xx = x
        # Restrict to 0 <= x <= 1/sqrt(2) with
        # acos(x) = pi/4  + acos(-x + sqrt(1-x^2))
        extra2 = 0
        rroot2 = longfloat(2).rsqrt()
        if x > rroot2:
            extra2 = 0.25*pi
            #x = (x - ((1+x).sqrt())*((1-x).sqrt()))*rroot2
            x = (x - (1-x*x).sqrt())*rroot2

        asinx = x.asin()
        #print ' asin arg ', float(x), float(asinx) - asin(float(x))
        acosx = pi2 - asinx
        #print ' acos arg ', float(x), float(acosx) - acos(float(x))
        if extra2:
            acosx = acosx - extra2
            #print ' acos xx ', float(xx), float(acosx) - acos(float(xx))
        
        acosx = extra + phase*acosx
        #print ' final ', float(xold), float(acosx) - acos(float(xold))

        return acosx

    def asin(self):
        '''
        asin(x) = sum(n) (-1)^n C(-1/2,n) x^(2n+1)/(2n+1)
        %       = x + (1/2)*x^3/3 + (1*3/2*4)*x^5/5 + (1*3*5/2*4*6)*x^7/7 ...

        asin(x) = -asin(-x)
        '''
        longfloat.nbits = longfloat.nbits + 8

        x = longfloat(self)
        one = longfloat(1)
        if (x < -one) or (x > one):
            longfloat.nbits = longfloat.nbits - 8
            raise ValueError, 'longfloat asin argument not in [-1,1]'

        input_phase = 1
        if x < 0:
            input_phase = -1
            x = -x

        pi  = self.pi()
        pi2 = pi*0.5
        if x == one:
            longfloat.nbits = longfloat.nbits - 8
            return input_phase*pi2

        # Shift the argument into [0,1/sqrt(2)] using
        # asin(x) = pi/2 - asin(sqrt(1-x^2))  0<=x<=1
        # asin(x) =-pi/2 + asin(sqrt(1-x^2)) -1<=x<=0
        # We need y = sqrt(1-x^2) accurate enough so that asin(y)
        # is accurate compared to pi/2.

        extra = 0
        phase = 1
        if x > longfloat(2).rsqrt():
            extra = pi2
            phase = -1
            #x = ((1+x).sqrt())*((1-x).sqrt())
            x = (1-x*x).sqrt()

        # Shift the argument into [0,1/2] using
        # asin(x) = pi/6 + asin(x*sqrt(3)/2 - sqrt(1-x^2)/2)

        didpi6 = 0
        half = longfloat((-1,1))
        if x > half:
            didpi6 = 1
            #x = (x*longfloat(3).sqrt() - ((1+x).sqrt())*((1-x).sqrt())) * half
            x = (x*longfloat(3).sqrt() - (1-x*x).sqrt()) * half
            #print ' shifting x to ', x, half
                 
        # asin(x) = sum(n) (-1)^n C(-1/2,n) x^(2n+1)/(2n+1)
        # %       = x + (1/2)*x^3/3 + (1*3/2*4)*x^5/5 + (1*3*5/2*4*6)*x^7/7 ...

        x2 = x*x
        term = x
        result = longfloat(0)
        test = longfloat((-self.nbits,1))

        # 0 <= x2 <= 0.25 so we get at least 2 bits per iteration
        maxi = self.nbits + 16
        for i in range(0,maxi,2):
            result = result + term / (i+1)
            term = term*x2*(i+1) / (i+2)
            if abs(term) < test:
                break
        if i >= (maxi-2):
            raise ArithmeticError, "longfloat asin did not converge"

        if didpi6:
            result = result + pi/6
            
        if phase < 0:
            result = -result
        result = extra + result

        longfloat.nbits = longfloat.nbits - 8
        result.truncate()
        if input_phase < 0:
            result = -result
        
        return result

    def atan(x):
        '''
        atan(x) = sum(n) (-1)^n x^(2n+1) / (2n+1)
        %       = x - x^3/3 + x^5/5 - x^7/7      |x| < 1

        atan(x) = - atan(-x) for all x

        atan(x) = pi/2 - atan(1/x) for x >= 0

        atan(x) = pi/2 - acos(x/sqrt(x^2+1)) for all x

        atan(x) = asin(x/sqrt(x^2+1)) for all x

        atan(x) = acot(-x) - pi/2 = pi/2 - acot(x) for all x

        atan(x) + atan(y) = atan((x+y)/(1-x*y))      
        '''

    def atan2(self,x,y):
        '''
        atan(y/x) with apropriate phase
        '''

    def eps(self):
        '''
        Returns eps the smallest number such that 1+eps > 1.
        '''
        eps = longfloat((-longfloat.nbits,1))

        #print ' eps   ', repr(eps)
        #print ' eps   ', eps
        #print ' 1+eps ', 1+eps
        #print ' 1+eps/2', 1+longfloat((-longfloat.nbits-1,1))
        #print ' 1+eps/4', 1+longfloat((-longfloat.nbits-2,1))

        return eps
        
    def __nbits_in_mantissa(self):
        ''' Return the no. of bits in the mantissa '''
        
        absm = abs(self.mantissa)
        if absm == 0:
            return 0

        # Find bounds
        klo = khi = 2*self.nbits+2
        if (absm >> khi):
            khi = khi + 1
            while (absm >> khi):
                klo = khi
                khi = khi << 1
        else:
            klo = klo - 2
            while not (absm >> klo):
                khi = klo
                klo = klo >> 1
            
        # Binary search
        delta = (khi - klo) >> 1
        while delta > 0:
            k = klo + delta
            if (absm >> k):
                klo = k
            else:
                khi = k
            delta = (khi - klo) >> 1
        return klo
        
    def truncate(self):
        '''
        Truncates self to the required number of significant binary digits.
        '''
        nchopped = self.__nbits_in_mantissa() - longfloat.nbits
        if nchopped > 0:
            round = (self.mantissa > 0) and ((self.mantissa >> (nchopped-1)) & 1L)
            self.mantissa = self.mantissa >> nchopped
            if round: self.mantissa = self.mantissa + 1
            self.exponent = self.exponent + nchopped
            if not self.mantissa: self.exponent = 0
 
    def other_truncate(self, nbits=None):
        '''
        Truncates self to the required number of significant binary digits.
        '''
        if nbits == None:
            nbits = self.nbits
        
        eold = e = self.exponent
        m = self.mantissa

        k = nbits  # Multiplications typically double the no. of bits
        test  = (1L << nbits)-1  # The biggest permisible mantissa

        flag = (abs(m) > test)
        while flag:
            mk = m >> k
            flag2 = (abs(mk) > test)
            if (k == 1) or flag2:
                m = mk
                flag = flag2
                e = e + k
                if k > 1:
                    k = k >> 1
            else:
                k = k >> 1

        if self.debug:
            print 'longfloat: truncate: nbits lost =', e - eold
        if m == 0L: e = 0

        # Comment out the next three lines to stop rounding
        if (e-eold):
            if (self.mantissa >> (e-eold-1)) & 1:
                m = m + 1

        self.exponent = e
        self.mantissa = m

    def __div__(self, other):
        return self*other.reciprocal()

    def __rdiv__(self,other):
        return other*self.reciprocal()
        

def __mp_do_test(test_nbits,ntest=10,ntime=10):
    from whrandom import random, seed
    from sys import exit
    from time import clock

    seed((3*test_nbits-1)%256,(5*test_nbits-2)%256,(7*test_nbits-3)%256)

    savenbits = longfloat.nbits
    longfloat.nbits = test_nbits
    print '\n\n Testing extended precision routines with nbits =', \
          longfloat.nbits
    print ' '
    
    def nbits(eps):
        return int(-log(abs(eps))/log(2.0))

    fac = longfloat(1)/longfloat(3)
    
    # Test reciprocal
    nbits_worst = longfloat.nbits
    n = ntest
    print ' n =',n
    maxx = 0
    minx = minabsx = 1e300
    while n > 0:
        x = longfloat((random()-0.5)*10.0)
        x = x*x*x*x*x # To fill out the number of bits
        if x:
            maxx = max(x,maxx)
            minx = min(x,minx)
            minabsx = min(abs(x),minabsx)
            eps = abs(float(x/x - 1.0))
            if eps:
                nbits_worst = min(nbits_worst, nbits(eps))
        n = n - 1
    print (' maxx=%+9.1e minx=%+9.1e minabsx=%+9.1e ' % (maxx, minx, minabsx))
    print ' Reciprocal:  eps = (x(1/x)-1)       :  worst nbits = ', nbits_worst

    x = longfloat(33)
    start = clock()
    n = ntime
    while n > 0:
        x.reciprocal()
        n = n - 1
    used = (clock() - start)/ntime
    print ' Time per call ', used
    print ' '
    
    # Test sqrt
    nbits_worst = longfloat.nbits
    n = ntest
    print ' n =',n
    maxx = 0
    minx = minabsx = 1e300
    while n > 0:
        x = longfloat(random()*10.0)
        x = x*x*x*x*x # To fill out the number of bits
        if x:
            maxx = max(x,maxx)
            minx = min(x,minx)
            minabsx = min(abs(x),minabsx)
            y = x.sqrt()
            # If there is a relative error eps in y=sqrt(x) then
            # eps = (y*y - x)/(2x)
            eps = 0.5*abs(float((y*y - x))/float(x))
            if eps:
                nbits_worst = min(nbits_worst, nbits(eps))
        n = n - 1
    print (' maxx=%+9.1e minx=%+9.1e minabsx=%+9.1e ' % (maxx, minx, minabsx))
    print ' Square root: eps = (sqrt(x)^2-x)/2x :  worst nbits = ', nbits_worst
    x = longfloat(33)
    start = clock()
    n = ntime
    while n > 0:
        x.rsqrt()
        n = n - 1
    used = (clock() - start)/ntime
    print ' Time per call ', used
    print ' '
    
    # Test log(exp)
    nbits_worst = longfloat.nbits
    n = ntest
    print ' n =',n
    maxx = 0
    minx = minabsx = 1e300
    while n > 0:
        x = longfloat((random()-0.5)*10.0)
        x = x*x*x*x*x # To fill out the number of bits
        if x:
            maxx = max(x,maxx)
            minx = min(x,minx)
            minabsx = min(abs(x),minabsx)
            y = x.exp()
            eps = abs(float((y.log() - x))/float(x))
            if eps:
                # If there is a relative error eps1 in log(x)
                # and eps2 in exp(x), then the relative error in
                # -- exp(log(x)) is eps2 + eps1*|logx|
                # -- log(exp(x)) is eps1 + eps2/|x|
                absx = abs(float(x))
                eps = eps /(1.0 + 1.0/absx)
                nbits_worst = min(nbits_worst, nbits(eps))
        n = n - 1
    print (' maxx=%+9.1e minx=%+9.1e minabsx=%+9.1e ' % (maxx, minx, minabsx))
    print ' log & exp :  eps = (log(exp(x))-x)/x:  worst nbits = ', nbits_worst
    x = longfloat(33)
    scale = longfloat(77)
    start = clock()
    n = ntime
    while n > 0:
        x.log()
        n = n - 1
        x = x * scale
    used = (clock() - start)/ntime
    print ' Time per call for log ', used
    x = longfloat(3)
    start = clock()
    n = ntime
    while n > 0:
        x.exp()
        n = n - 1
    used = (clock() - start)/ntime
    print ' Time per call for exp ', used
    print ' '

    # Test exp(log)
    nbits_worst = longfloat.nbits
    n = ntest
    print ' n =',n
    maxx = 0
    minx = minabsx = 1e300
    while n > 0:
        x = longfloat(random()*10.0)
        x = x*x*x*x*x # To fill out the number of bits
        if x:
            maxx = max(x,maxx)
            minx = min(x,minx)
            minabsx = min(abs(x),minabsx)
            y = x.log()
            eps = abs(float((y.exp() - x))/float(x))
            if eps:
                abslogx = abs(float(y))
                # If there is a relative error eps1 in log(x)
                # and eps2 in exp(x), then the relative error in
                # -- exp(log(x)) is eps2 + eps1*|logx|
                # -- log(exp(x)) is eps1 + eps2/|x|
                eps = eps / (1.0 + abslogx)
                nbits_worst = min(nbits_worst, nbits(eps))
        n = n - 1
    print (' maxx=%+9.1e minx=%+9.1e minabsx=%+9.1e ' % (maxx, minx, minabsx))
    print ' log & exp :  eps = (exp(log(x))-x)/x:  worst nbits = ', nbits_worst
    print ' '

    # Test sine and asin
    # Note that asin(sin(x)) is inherently unstable so test just sin(asin(x))
    nbits_worst = longfloat.nbits
    n = ntest
    print ' n =',n
    maxx = 0
    minx = minabsx = 1e300
    xworst = 0.0
    while n > 0:
        x = longfloat((random()-0.5))*2.0  # x in [-1,1]
        if x:
            maxx = max(x,maxx)
            minx = min(x,minx)
            minabsx = min(abs(x),minabsx)
            eps = float(x.asin().sin() - x)/float(x)
            if eps:
                nbb = nbits(eps)
                if nbb < nbits_worst:
                    xworst = x
                    nbits_worst = nbb
                #nbits_worst = min(nbits_worst, nbits(eps))
                #print float(x), nbits(eps)
        n = n - 1
    print (' maxx=%+9.1e minx=%+9.1e minabsx=%+9.1e ' % (maxx, minx, minabsx))
    print ' sin       : eps = (sin(asin(x))-x)/x:  worst nbits = ', nbits_worst
    print ' xworst ', xworst
    x = longfloat(2.7)
    start = clock()
    n = ntime
    while n > 0:
        x.sin()
        n = n - 1
    used = (clock() - start)/ntime
    print ' Time per call for sin  ', used
    x = longfloat(0.33333)
    start = clock()
    n = ntime
    while n > 0:
        x.asin()
        n = n - 1
    used = (clock() - start)/ntime
    print ' Time per call for asin ', used

    
    # Test cosine and acos
    nbits_worst = longfloat.nbits
    n = ntest
    print ' n =',n
    maxx = 0
    minx = minabsx = 1e300
    xworst = 0.0
    while n > 0:
        x = longfloat((random()-0.5))*2.0  # x in [-1,1]
        if x:
            maxx = max(x,maxx)
            minx = min(x,minx)
            minabsx = min(abs(x),minabsx)
            eps = float(x.acos().cos() - x)/float(x)
            if eps:
                nbb = nbits(eps)
                if nbb < nbits_worst:
                    xworst = x
                    nbits_worst = nbb
        n = n - 1
    print (' maxx=%+9.1e minx=%+9.1e minabsx=%+9.1e ' % (maxx, minx, minabsx))
    print ' cos       : eps = (cos(acos(x))-x)/x:  worst nbits = ', nbits_worst
    print ' xworst ', xworst
    x = longfloat(2.7)
    start = clock()
    n = ntime
    while n > 0:
        x.cos()
        n = n - 1
    used = (clock() - start)/ntime
    print ' Time per call for cos  ', used
    x = longfloat(0.33333)
    start = clock()
    n = ntime
    while n > 0:
        x.acos()
        n = n - 1
    used = (clock() - start)/ntime
    print ' Time per call for acos ', used

    

    # Test e and pi
    maplee  = '2.718281828459045235360287471352662497757247093699959574966967627724076630353547594571382178525166427427466391932003059921817413596629043572900334295260595630738132328627943490763233829880753195251019011573834187930702154089149934884167509244761460668082264800168477411853742345442437107539077744992069551702761838606261331384583000752044933826560297606737113200709328709127443747047230696977209310141692836819025515108657463772111252389784425056953696770785449969967946864454905987931636889230098793127736178215424999229576351482208269895193668033182528869398496465105820939239829488793320362509443117301238197068416140397019837679320683282376464804295311802328782509819455815301756717361332069811250996181881593041690351598888519345807273866738589422879228499892086805825749279610484198444363463244968487560233624827041978623209002160990235304369941849146314093431738143640546253152096183690888707016768396424378140592714563549061303107208510383750510115747704171898610687396965521267154688957035035'
    maplepi = '3.141592653589793238462643383279502884197169399375105820974944592307816406286208998628034825342117067982148086513282306647093844609550582231725359408128481117450284102701938521105559644622948954930381964428810975665933446128475648233786783165271201909145648566923460348610454326648213393607260249141273724587006606315588174881520920962829254091715364367892590360011330530548820466521384146951941511609433057270365759591953092186117381932611793105118548074462379962749567351885752724891227938183011949129833673362440656643086021394946395224737190702179860943702770539217176293176752384674818467669405132000568127145263560827785771342757789609173637178721468440901224953430146549585371050792279689258923542019956112129021960864034418159813629774771309960518707211349999998372978049951059731732816096318595024459455346908302642522308253344685035261931188171010003137838752886587533208381420617177669147303598253490428755468731159562863882353787593751957781857780532171226806613001927876611195909216420199'
    ndigit = int(longfloat.nbits/(log(10)/log(2)))
    ndigit = min(1001,ndigit+2)
    
    print (' The error in e and pi should be in the last d.p. (about %9.1e)' %
           (3.0*(0.5**longfloat.nbits)))
    print '    Computed e   ', longfloat(1).exp()
    print '       Maple e    ', maplee[:ndigit]
    print '    Computed pi  ', longfloat(1).pi()
    print '       Maple pi   ', maplepi[:ndigit]
    print ' ' 
    print ' nbits at end ', longfloat.nbits
    print ' '
    
    longfloat.nbits = savenbits
    print ' restored longfloat.nbits ', savenbits

if __name__ == "__main__":
    from time import clock
    I = interpolators()
##     print ' EXP '
##     vv = I.exp(longfloat(0.25))
##     start = clock()
##     for i in range(10):
##         vv = I.exp(longfloat(0.25))
##     used = (clock() - start)/10
##     print ' NEW ', used
##     start = clock()
##     for i in range(10):
##         vv = longfloat(0.25).exp()
##     used = (clock() - start)/10
##     print ' OLD ', used
    ## print ' SIN '
##     print I.sin(0.25,1e-16)
##     print sin(0.25)
##     print ' COS '
##     print I.cos(0.25,1e-16)
##     print cos(0.25)
    print ' LOG '
    print I.log(longfloat(0.9))
    print longfloat(0.9).log()
    q = longfloat(0.72)
    I.log(q)
##     start = clock()
##     for i in range(10):
##         vv = I.log(q)
##     used = (clock() - start)/10
##     print ' NEW ', used
##     start = clock()
##     for i in range(10):
##         vv = q.log()
##     used = (clock() - start)/10
##     print ' OLD ', used
    import profile
    profile.run('for i in range(100): q.reciprocal()')
##     print ' ASIN by inversion'
##     print I.invert(0.25,I.sindsin,0.3,1e-16)
##     print asin(0.25)
##     print ' ACOS by inversion'
##     print I.invert(0.25,I.cosdcos,1.1,1e-16)
##     print acos(0.25)
##     print ' EXP=ALOG by inversion'
##     print I.invert(0.25,I.logdlog,1.0,1e-16)
##     print exp(0.25)
##     print ' LOG=AEXP by inversion'
##     print I.invert(1.35,I.expdexp,0.2,1e-16)
##     print log(1.35)
    #print ' HI'
    #__mp_do_test(104,ntest=3,ntime=3)
    #__mp_do_test(208,ntest=3,ntime=3)
    #__mp_do_test(500)
    #__mp_do_test(1000)
else:
    #print ' reimported mp'
    pass

## def sqrt(x):
##     return longfloat(x).sqrt()

## def sin(x):
##     return longfloat(x).sin()

## def cos(x):
##     return longfloat(x).cos()

## def asin(x):
##     return longfloat(x).asin()

## def acos(x):
##     return longfloat(x).acos()

## def exp(x):
##     return longfloat(x).exp()

## def log(x):
##     return longfloat(x).log()




# TODO: MPI_Barrier(MPI_COMM_WORLD) around printf's
def _header(complex_c, complex_a, complex_b):
    ret = """
#include <stdio.h>
#include <hwi/include/bqc/A2_inlines.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <unistd.h>
#include <pthread.h>
#include <mpi.h>
#include <limits.h>
#include <complex>

#define NAPTIME 3
#define MAX_POSIX_THREADS 64

typedef struct _pt_data {
    int myid;
    int total_threads;
    uint64_t _tc;
    struct _pt_data **args;
    uint64_t tt;
} pt_data;

static pthread_t thread_pool[MAX_POSIX_THREADS];
pt_data *thread_args[MAX_POSIX_THREADS];

static int mpi_size, mpi_rank;
static int num_posix_threads;
pthread_barrier_t barr;

void ran_fill(int n, double *a) {
    while (n--) *a++ = ((double)rand())/RAND_MAX;
}

inline void start_timer(pt_data *pt) {
    pt->_tc = GetTimeBase();
}

inline uint64_t stop_timer(pt_data *pt) {
    return GetTimeBase() - pt->_tc;
}

void mtxm(long dimi, long dimj, long dimk, long extb,
""" + "double {} * c, const double {} * a, const double {} * b) {{".format(complex_c, complex_a, complex_b) + """
    int i, j, k;
    for (k=0; k<dimk; ++k) {
        for (j=0; j<dimj; ++j) {
            for (i=0; i<dimi; ++i) {
                c[i*dimj+j] += a[k*dimi+i]*b[k*extb+j];
            }
        }
    }
}
"""
    return ret

def _tester(i, m, trans=False):
    ret = """
    if (mpi_rank=={i}) {{
        for (loop=0; loop<30; ++loop) {{
            rc = pthread_barrier_wait(&barr);
            if(rc != 0 && rc != PTHREAD_BARRIER_SERIAL_THREAD) {{
                printf("Could not wait on barrier\\n");
                exit(-1);
            }}

            start_timer(pt);
            {m}(ni,nj,nk,nj,c,a,b);"""
    if trans:
        ret += """
            {m}(ni,nj,nk,nj,a,c,b);
            {m}(ni,nj,nk,nj,c,a,b);"""
    ret += """
            uint64_t tt = stop_timer(pt);
            pt->tt += tt;
        }}
        pt->tt /= 30;

        rc = pthread_barrier_wait(&barr);
        if(rc != 0 && rc != PTHREAD_BARRIER_SERIAL_THREAD) {{
            printf("Could not wait on barrier\\n");
            exit(-1);
        }}

        if (pt->myid == 0) {{
            uint64_t maxtt = 0; // longest time
            uint64_t mintt = ULONG_MAX; // shortest time
            uint64_t sumtt = 0;

            int i;
            for (i=0; i<pt->total_threads; i++) {{
                uint64_t itt;
                itt = pt->args[i]->tt;
                maxtt = maxtt > itt ? maxtt : itt;
                mintt = mintt < itt ? mintt : itt;
                sumtt += itt;
            }}

            double avgtt = (double)sumtt/(double)i;
            printf("{m} [%2d pthreads]: %20s %3ld %3ld %3ld %8.2f (avg: %8.2f min: %llu max: %llu)\\n", pt->total_threads, s, ni, nj, nk, nflop/maxtt, avgtt, mintt, maxtt);
        }}

        rc = pthread_barrier_wait(&barr);
        if(rc != 0 && rc != PTHREAD_BARRIER_SERIAL_THREAD) {{
            printf("Could not wait on barrier\\n");
            exit(-1);
        }}
    }}"""
    return ret.format(i=i, m=m)

def _timer(mtxms, complex_c, complex_a, complex_b):
    ret = """void timer(pt_data *pt, const char* s, long ni, long nj, long nk, double {} *a, double {} *b, double {} *c) {{
    double fastest = 0.0;
    double nflop = 2.0{}{}*ni*nj*nk*(pt->total_threads);
    long loop;
    int rc;
    pt->tt = 0;
""".format(complex_a, complex_b, complex_c, complex_a and "*2.0", complex_b and "*2.0")

    for i, m in enumerate(mtxms):
        ret += _tester(i, m)

    ret += "}\n"
    return ret

def _transtimer(mtxms, complex_c, complex_a, complex_b):
    ret = """void trantimer(pt_data *pt, const char* s, long ni, long nj, long nk, double {} *a, double {} *b, double {} *c) {{
    double fastest = 0.0;
    double nflop = 3.0*2.0{}{}*ni*nj*nk*(pt->total_threads);
    long loop;
    int rc;
    pt->tt = 0;
""".format(complex_a, complex_b, complex_c, complex_a and "*2.0", complex_b and "*2.0")

    for i, m in enumerate(mtxms):
        ret += _tester(i, m, True)

    ret += "}\n"
    return ret

def _main(mtxms, complex_c, complex_a, complex_b):
    ret = """void* entry(void* arg) {{
    pt_data *pt = (pt_data*)arg;

    const long nimax=30*30;
    const long njmax=100;
    const long nkmax=100;
    long ni, nj, nk, i, m;

    double {ca} *a;
    double {cb} *b;
    double {cc} *c;
    double {cc} *d;

    posix_memalign((void **) &a, 32, nkmax*nimax*sizeof(double {ca}));
    posix_memalign((void **) &b, 32, nkmax*njmax*sizeof(double {cb}));
    posix_memalign((void **) &c, 32, nimax*njmax*sizeof(double {cc}));
    posix_memalign((void **) &d, 32, nimax*njmax*sizeof(double {cc}));

    ran_fill(nkmax*nimax{cax}, (double*)a);
    ran_fill(nkmax*njmax{cbx}, (double*)b);

    if (pt->total_threads == 1) {{
        for (ni=2; ni<60; ni+=2) {{
            for (nj=2; nj<100; nj+=6) {{
                for (nk=2; nk<100; nk+=6) {{
                    for (i=0; i<ni*nj; ++i) d[i] = c[i] = 0.0;
                    mtxm(ni,nj,nk,nj,c,a,b);
"""
    for i, f in enumerate(mtxms):
        ret += "                    if (mpi_rank=={}) {}(ni,nj,nk,nj,d,a,b);\n".format(i, f)
    ret += """
                    for (i=0; i<ni*nj; ++i) {{
                        double err = abs(d[i]-c[i]);
                        if (err > 1e-{error}) {{
"""
    for i, f in enumerate(mtxms):
        ret += """                            if (mpi_rank=={}) fprintf(stderr, "{} error %d: %ld %ld %ld %e\\n",mpi_rank,ni,nj,nk,err);\n""".format(i, f)
    ret += """
                        }}
                    }}
                }}
            }}
        }}
    }}

    for (ni=2; ni<60; ni+=2) timer(pt, "(m*m)T*(m*m)", ni,ni,ni,a,b,c);
    for ( m=2; m<=30;  m+=2) timer(pt, "(m*m,m)T*(m*m)", m*m,m,m,a,b,c);
"""
    if (complex_a and complex_b) or (not complex_a and not complex_b):
        ret += """    for ( m=2; m<=30;  m+=2) trantimer(pt, "tran(m,m,m)", m*m,m,m,a,b,c);\n"""
    ret += """
    for ( m=2; m<=20;  m+=2) timer(pt, "(20*20,20)T*(20,m)", 20*20,m,20,a,b,c);

    free(a);
    free(b);
    free(c);
    free(d);

    return NULL;
}}

int main(int argc, char *argv[]) {{
    int i, rc;
    int provided;

    MPI_Init_thread(&argc,&argv,MPI_THREAD_MULTIPLE,&provided);
    assert(provided == MPI_THREAD_MULTIPLE);

    MPI_Comm_size(MPI_COMM_WORLD,&mpi_size);
    MPI_Comm_rank(MPI_COMM_WORLD,&mpi_rank);

    srand(argc > 42 ? argc : 1);

    for (num_posix_threads = 1; num_posix_threads <= MAX_POSIX_THREADS; num_posix_threads*=2) {{
        MPI_Barrier(MPI_COMM_WORLD);

        for (i=0; i<num_posix_threads; i++) {{
            thread_args[i] = (pt_data*)malloc(sizeof(pt_data));
            thread_args[i]->myid = i;
            thread_args[i]->total_threads = num_posix_threads;
            thread_args[i]->args = thread_args;
        }}

        if(pthread_barrier_init(&barr, NULL, num_posix_threads)) {{
            printf("Could not create a barrier\\n");
            return -1;
        }}

        for (i=0 ; i<num_posix_threads ; i++) {{
            rc = pthread_create(&thread_pool[i], NULL, entry, thread_args[i]);
            assert(rc==0);
        }}

        sleep(NAPTIME);

        for (i=0 ; i<num_posix_threads ; i++) {{
            rc = pthread_join(thread_pool[i],NULL);
            assert(rc==0);
        }}

        MPI_Barrier(MPI_COMM_WORLD);

        for (i=0; i<num_posix_threads; i++) {{
            free(thread_args[i]);
        }}
    }}

    MPI_Barrier(MPI_COMM_WORLD);

    MPI_Finalize();

    return 0;
}}"""
    return ret.format(ca=complex_a, cb=complex_b, cc=complex_c, 
            cax=(complex_a and "*2"), cbx=(complex_b and "*2"), 
            error=(complex_c and "12" or "14"))

def tester_gen_bgq(f, mtxmspecs, mtxm_gen, complex_a, complex_b):
    r = lambda x: x.replace("double complex", "__complex__ double")
    complex_c = (complex_a or complex_b) and "complex" or ""
    complex_a = complex_a and "complex" or ""
    complex_b = complex_b and "complex" or ""
    print(r(_header(complex_c, complex_a, complex_b)), file=f)
    for mtxm in mtxmspecs:
        mtxm_gen(f, *mtxm)
    mtxms = [x[-1] for x in mtxmspecs]
    print(r(_timer(mtxms, complex_c, complex_a, complex_b)), file=f)
    if not (bool(complex_a) ^ bool(complex_b)):
        print(r(_transtimer(mtxms, complex_c, complex_a, complex_b)), file=f)
    print(r(_main(mtxms, complex_c, complex_a, complex_b)), file=f)



import sys

basis_name = "6-31G"

print '<?xml version="1.0" ?>'
print '<name>'
print "  ", basis_name
print '</name>'

while 1:
    line = sys.stdin.readline()
    if not line: break

    if line[0:5] == "basis":
        symbol = line.split()[1].split("_")[0][1:]
        print '<basis symbol="%s">' % symbol

        line = sys.stdin.readline().split()
        while 1:
            if line[0] == "end":
                break
            shelltype = line[1]
            coeff = []
            pcoeff = []
            expnt = []
            while 1:
                line = sys.stdin.readline().split()
                e,c,p = 0.0,0.0,0.0
                try:
                    e,c = float(line[0]), float(line[1])
                    if shelltype == 'SP': p = float(line[2])
                    expnt.append(e)
                    coeff.append(c)
                    pcoeff.append(p)
                except:
                    break
            nprim = len(expnt)
            if shelltype == "SP": shelltype = "L"
            print '  <shell type="%s" nprim="%d">' % (shelltype, nprim)
            print '    <exponents>'
            for e in expnt: print "      %.8f" % e
            print '    </exponents>'
            if shelltype == 'L':
                print '    <scoefficients>'
                for e in coeff: print "      %.8f" % e
                print '    </scoefficients>'
                print '    <pcoefficients>'
                for e in pcoeff: print "      %.8f" % e
                print '    </pcoefficients>'
            else:
                print '    <coefficients>'
                for e in coeff: print "      %.8f" % e
                print '    </coefficients>'

            print '  </shell>'

        print '</basis>'




#!/usr/bin/env python3

import fileinput

class COLORS:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'

    def disable(self):
        self.HEADER = ''
        self.OKBLUE = ''
        self.OKGREEN = ''
        self.WARNING = ''
        self.FAIL = ''
        self.ENDC = ''

# [ (variant, pthreads, (shape, i, j, k), gflops), ]

def get_data():
    results = []
    variants = set()
    tests = set()
    pthreads = set()
    for line in fileinput.input():
        line = line.strip().replace('[', '').replace('pthreads]:', '').replace('(avg', 'avg').replace(':', '')[:-1].split()[:-6]
        test = tuple([line[2]] + [int(x) for x in line[3:-1]])
        gflops = float(line[-1])
        results.append((line[0], int(line[1]), test, gflops))
        variants.add(line[0])
        tests.add(test)
        pthreads.add(int(line[1]))
    return (results, sorted(variants), sorted(pthreads), sorted(tests))

def print_best_per_benchmark(results):
    best_avg, blas_avg, imp_avg = 0, 0, 0
    for test in sorted(results):
        best = max(results[test], key=lambda x: x[1])
        best_avg += best[1]
        color = COLORS.OKGREEN
        print(color + '{:40}'.format(' '.join([str(x) for x in test])) + COLORS.ENDC, '{:40}{:6.2f}'.format(*best))
    best_avg /= len(results)
    print('{:40} {:40}{:6.2f}'.format('', 'average', best_avg))

def print_top_n_variants(results, n=10):
    variants = {}
    for test in results:
        for variant, gflops in results[test]:
            if variant not in variants:
                variants[variant] = 0
            variants[variant] += gflops / len(results)

    topn = sorted(variants, key=lambda x: variants[x], reverse=True)[:n]
    for x in topn:
        print('{:40}{:6.2f}'.format(x, variants[x]))


if __name__ == '__main__':
    results, variants, pthreads, tests = get_data()
    for pc in pthreads:
        pcfilter = list(filter(lambda x: x[1] == pc, results))
        for test in tests:
            tfilter = sorted(filter(lambda x: x[2] == test, pcfilter), key=lambda x: x[3])
            if tfilter:
                print(pc, test, tfilter[-1][0], tfilter[-1][3])
        avgs = {}
        for variant in variants:
            vfilter = list(filter(lambda x: x[0] == variant, pcfilter))
            avg = sum(x[3] for x in vfilter)/len(vfilter)
            avgs[variant] = avg
        top = sorted(avgs, key=lambda x: avgs[x])[-5:][::-1]
        for v in top:
            print(pc, v, avgs[v])


    #print(COLORS.HEADER + "Best per benchmark:" + COLORS.ENDC)
    #print_best_per_benchmark(results)
    #print(COLORS.HEADER + "Top 10 Varaints:" + COLORS.ENDC)
    #print_top_n_variants(results)
